#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <ctype.h>
#include <time.h>
#include "utilities.h"
#include "Ring.h"

//  -----------------------------------------------
void coord_st_To_xy(parameters par, int idom, int i, int j, double s, double t, derivs_1D G,
                    coordinate *x, coordinate *y)
{// Calculation of the coordinates (x,y) from (s,t) along with first and second
 // derivatives with repect to s and t
	switch(idom){
		case 0: coord_0(par, i, j, s, t, G, x, y); break;
		case 1: coord_1(par, i, j, s, t, G, x, y); break;
		case 2: coord_2(par, i, j, s, t, G, x, y); break;
		case 3: coord_3(par, i, j, s, t, G, x, y); break;
	}
}
//  -----------------------------------------------
void coord_st_To_xy_all_st(parameters par, int idom, int i, int j, double s, double t,
                            derivs_1D Gcheb, coordinate *x, coordinate *y)
{// Calculation of the coordinates (x,y) from (s,t) along with first and second
 // derivatives with repect to s and t
	switch(idom){
		case 0: coord_0(par, i, j, s, t, Gcheb, x, y); break;
		case 1: coord_1(par, i, j, s, t, Gcheb, x, y); break;
		case 2: coord_2_all_st(par, i, j, s, t, Gcheb, x, y); break;
		case 3: coord_3_all_st(par, s, t, Gcheb, x, y); break;
	}
}
//  -----------------------------------------------
void Dcoord_st_To_xy(parameters par, parameters Dpar, int idom, int i, int j, double s, double t, derivs_1D G, derivs_1D DG, coordinate *x, coordinate *y, coordinate *Dx, coordinate *Dy)
{// Linear version of 'coord_st_To_xy'
	switch(idom){
		case 0: Dcoord_0(par, Dpar, i, j, s, t, G, DG, x, y, Dx, Dy); break;
		case 1: Dcoord_1(par, Dpar, i, j, s, t, G, DG, x, y, Dx, Dy); break;
		case 2: Dcoord_2(par, Dpar, i, j, s, t, G, DG, x, y, Dx, Dy); break;
		case 3: Dcoord_3(par, Dpar, i, j, s, t, G, DG, x, y, Dx, Dy); break;
	}
}
//   -----------------------------------------------
double Func_xm(parameters par)
{// Calculation of xm = rm^2
	double
		xi = par.xi,       xo = par.xo,
		ri = sqrt(xi),     ro = sqrt(xo),
		rm = 0.5*(ri+ro),  xm = rm*rm;
	
	return xm;
}
//   -----------------------------------------------
double Func_Dxm(parameters par, parameters Dpar)
{// Linear version of 'Func_xm'
	double
		xi = par.xi,       xo = par.xo,    Dxi = Dpar.xi,    Dxo = Dpar.xo,
		ri = sqrt(xi),     ro = sqrt(xo),  Dri = 0.5*Dxi/ri, Dro = 0.5*Dxo/ro,
		rm = 0.5*(ri+ro), Drm = 0.5*(Dri+Dro),               Dxm = 2.*rm*Drm;
	
	return Dxm;
}
//   -----------------------------------------------
void xys_To_xy(double xm, coordinate xs, coordinate ys, coordinate *x, coordinate *y)
{// calculation of x(xs,ys), y(xs,ys), cf. eqns. (8), (9)
	coordinate a, b, c, f1, f2;

	a   = const_times_coord(xm, ys); 	// a  = xm*ys
	b   = const_plus_coord(1., ys);  	// b  = 1+ys
	c   = coordSum(xs, ys);			 	// c  = xs+ys
	f1  = coordQuotient(a, c);		 	// f1 = a/c = xm*ys/(xs+ys)
	f2  = coordQuotient(b, c);		 	// f2 = b/c = (1+ys)/(xs+ys)
	*x  = coordProduct(f1, f2);		 	// x  = f1*f2 = xm*ys*(1+ys)/(xs+ys)^2

	a   = const_times_coord(-xm, xs); 	// a  = -xm*xs
	b   = const_plus_coord(-1., xs);  	// b  = -1+xs
	f1  = coordQuotient(a, c);		 	// f1 = a/c = -xm*xs/(xs+ys)
	f2  = coordQuotient(b, c);		 	// f2 = b/c = (-1+xs)/(xs+ys)
	*y  = coordProduct(f1, f2);		 	// y  = f1*f2 = xm*xs*(1-xs)/(xs+ys)^2
}
//   -----------------------------------------------
void D_xys_To_xy(double xm, double Dxm, coordinate xs, coordinate ys, coordinate Dxs, coordinate Dys,               coordinate *x, coordinate *y, coordinate *Dx, coordinate *Dy)
{// Linear version of 'xys_To_xy'
	coordinate a, b, c, f1, f2, Da, Db, Dc, Df1, Df2;

	a   = const_times_coord(xm, ys);   Da  = D_const_times_coord(xm, Dxm, ys, Dys); // a  = xm*ys
	b   = const_plus_coord(1., ys);    Db  = Dys;                                   // b  = 1+ys
	c   = coordSum(xs, ys);            Dc  = coordSum(Dxs, Dys);                    // c  = xs+ys
	f1  = coordQuotient(a, c);         Df1 = DcoordQuotient(a, c, Da, Dc);          // f1 = a/c
	f2  = coordQuotient(b, c);         Df2 = DcoordQuotient(b, c, Db, Dc);          // f2 = b/c
	*x  = coordProduct(f1, f2);        *Dx = DcoordProduct(f1, f2, Df1, Df2);       // x  = f1*f2

	a   = const_times_coord(-xm, xs);  Da  = D_const_times_coord(-xm, -Dxm, xs, Dxs);// a  = -xm*xs
	b   = const_plus_coord(-1., xs);   Db  = Dxs;                                    // b  = -1+xs
	f1  = coordQuotient(a, c);         Df1 = DcoordQuotient(a, c, Da, Dc);           // f1 = a/c
	f2  = coordQuotient(b, c);         Df2 = DcoordQuotient(b, c, Db, Dc);           // f2 = b/c
	*y  = coordProduct(f1, f2);        *Dy = DcoordProduct(f1, f2, Df1, Df2);        // y  = f1*f2
}
//   -----------------------------------------------
void xy_To_xys(double xm, coordinate x, coordinate y, coordinate *xs, coordinate *ys)
{// calculation of x(xs,ys), y(xs,ys)
	coordinate a, ay, b, c, d, root;

	a    = const_plus_coord(xm, x);                          // a  = xm+x;
	b    = const_times_coord(-1., const_plus_coord(-xm, x)); // b  = xm-x;
	ay   = coordProduct(a, y);                               // ay = a*y
	c    = coordSum(coordProduct(b, b), ay);                 // c  = (xm-x)^2+y*(x+xm)
	d    = coordSum(coordProduct(y, y), ay);                 // d  = y^2     +y*(x+xm)
	root = coordRoot(coordSum(c,d));                         // root = Sqrt[(xm-x)^2+2*y*(x+xm)+y^2]
	
	// xs = 0.5*[ 1+(xm-x-y)/root]; xm-x-y = b-y
	*xs  = const_times_coord(0.5, const_plus_coord(+1., coordQuotient(coordDiff(b,y), root)));
	// ys = 0.5*[-1+(x+xm+y)/root]; xm+x+y = a+y
	*ys  = const_times_coord(0.5, const_plus_coord(-1., coordQuotient(coordSum( a,y), root)));
}
//   -----------------------------------------------
void D_xy_To_xys(double xm, double Dxm, coordinate x, coordinate y, coordinate Dx, coordinate Dy,                  coordinate *xs, coordinate *ys, coordinate *Dxs, coordinate *Dys)
{// Linear version of 'xy_To_xys'
	coordinate a, ay, b, c, d, root2, root, Da, Day, Db, Dc, Dd, Droot;

	a     = const_plus_coord(xm, x);                            // a  = xm+x;
	Da    = const_plus_coord(Dxm, Dx);
	b     = const_times_coord(-1., const_plus_coord(-xm,  x));  // b  = xm-x;
	Db    = const_times_coord(-1., const_plus_coord(-Dxm,Dx));
	ay    = coordProduct(a, y);
	Day   = DcoordProduct(a, y, Da, Dy);                        // ay = a*y
	c     = coordSum(coordProduct(b, b), ay);                   // c  = b*b+a*y
	Dc    = coordSum(DcoordProduct(b, b, Db, Db), Day);
	d     = coordSum(coordProduct(y, y), ay);                   // d  = y*y+a*y
	Dd    = coordSum(DcoordProduct(y, y, Dy, Dy), Day);
	root2 = coordSum(c,d);
	root  = coordRoot(root2);                                   // root = Sqrt[c+d]
	Droot = DcoordRoot(root2, coordSum(Dc,Dd));
	
	// xs = 0.5*[ 1+(xm-x-y)/root]; xm-x-y = b-y
	*xs   = const_times_coord(0.5, const_plus_coord(+1., coordQuotient(coordDiff(b,y), root)));
	*Dxs  = const_times_coord(0.5, DcoordQuotient(coordDiff(b,y), root, coordDiff(Db,Dy), Droot));
	// ys = 0.5*[-1+(x+xm+y)/root]; xm+x+y = a+y
	*ys   = const_times_coord(0.5, const_plus_coord(-1., coordQuotient(coordSum( a,y), root)));
	*Dys  = const_times_coord(0.5, DcoordQuotient(coordSum(a,y),  root,  coordSum(Da,Dy), Droot));
}
//---------------------------------------------------------------
void coord_0(parameters par, int i, int j, double s, double t, derivs_1D G, coordinate *x, coordinate *y)
{// Coordinate transformation in domain 0
	double s2=s*s, tm1=t-1., x0=X0FACT*par.xo, xm=Func_xm(par);
	coordinate ax, ay, b, xs, ys;
	
	ax.d0  = -   xm*s2*tm1; // ax = xm*s^2*(1-t)
	ax.d1  = -2.*xm*s *tm1;
	ax.d2  = -   xm*s2;
	ax.d11 = -2.*xm   *tm1;
	ax.d12 = -2.*xm*s;
	ax.d22 = 0.;
	
	ay.d0  =     xm*s2*t;  // ay = xm*s^2*t
	ay.d1  =  2.*xm*s *t;
	ay.d2  =     xm*s2;
	ay.d11 =  2.*xm   *t;
	ay.d12 =  2.*xm*s;
	ay.d22 = 0.;
	
	b      = const_plus_coord(x0, coordDiff(ax,ay)); // b  = x0 + ax - ay
	xs     = coordQuotient(ax, b);                   // xs = ax/b
	ys     = coordQuotient(ay, b);                   // ys = ay/b
	
	xys_To_xy(xm, xs, ys, x, y);                     // note: x+y = r^2 = x0/s^2
}
//---------------------------------------------------------------
void Dcoord_0(parameters par, parameters Dpar, int i, int j, double s, double t,
              derivs_1D G, derivs_1D DG, coordinate *x, coordinate *y, coordinate *Dx, coordinate *Dy)
{// Linear version of 'coord_0'
	double s2=s*s, tm1=t-1., x0=X0FACT*par.xo, xm=Func_xm(par), Dxm=Func_Dxm(par, Dpar);
	coordinate ax, ay, b, xs, ys, Dax, Day, Db, Dxs, Dys;
	
	ax.d0  = -   xm*s2*tm1;  Dax.d0  = -   Dxm*s2*tm1; // ax = xm*s^2*(1-t)
	ax.d1  = -2.*xm*s *tm1;  Dax.d1  = -2.*Dxm*s *tm1;
	ax.d2  = -   xm*s2;      Dax.d2  = -   Dxm*s2;
	ax.d11 = -2.*xm   *tm1;  Dax.d11 = -2.*Dxm   *tm1;
	ax.d12 = -2.*xm*s;       Dax.d12 = -2.*Dxm*s;
	ax.d22 = 0.;             Dax.d22 = 0.;
	
	ay.d0  =     xm*s2*t;    Day.d0  =     Dxm*s2*t;   // ay = xm*s^2*t
	ay.d1  =  2.*xm*s *t;    Day.d1  =  2.*Dxm*s *t;
	ay.d2  =     xm*s2;      Day.d2  =     Dxm*s2;
	ay.d11 =  2.*xm   *t;    Day.d11 =  2.*Dxm   *t;
	ay.d12 =  2.*xm*s;       Day.d12 =  2.*Dxm*s;
	ay.d22 = 0.;             Day.d22 = 0.;
	
	b      = const_plus_coord(x0, coordDiff(ax,ay));  // b  = x0 + ax - ay
	Db     = coordDiff(Dax,Day);
	xs     = coordQuotient(ax, b);                    // xs = ax/b
	Dxs    = DcoordQuotient(ax, b, Dax, Db);
	ys     = coordQuotient(ay, b);                    // ys = ay/b
	Dys    = DcoordQuotient(ay, b, Day, Db);
	
	D_xys_To_xy(xm, Dxm, xs, ys, Dxs, Dys, x, y, Dx, Dy);
}
//---------------------------------------------------------------
void coord_1(parameters par, int i, int j, double s, double t, derivs_1D G, coordinate *x, coordinate *y)
{// Coordinate transformation in domain 1
	double x1=X1FACT*par.xi, x0=X0FACT*par.xo, xm=Func_xm(par);
	coordinate c_s, c_t, oms, omt, denom, xs, ys;

	c_s.d0 = s; c_s.d1 = 1.; c_s.d2 = c_s.d11 = c_s.d12 = c_s.d22 = 0.;
	c_t.d0 = t; c_t.d2 = 1.; c_t.d1 = c_t.d11 = c_t.d12 = c_t.d22 = 0.;
	oms.d0 = 1.-s; oms.d1 = -1.; oms.d2 = oms.d11 = oms.d12 = oms.d22 = 0.;
	omt.d0 = 1.-t; omt.d2 = -1.; omt.d1 = omt.d11 = omt.d12 = omt.d22 = 0.;

	denom.d0  = x0 + xm*(1.-2.*t); // denom = x0 + xm*(1-2t)
	denom.d2  = -2.*xm;
	denom.d1  = denom.d11 = denom.d12 = denom.d22 = 0.;

	xs = coordSum(coordQuotient(const_times_coord(xm, coordProduct(c_s, omt)), denom) , oms); // xs=xm*s*(1-t)/(denom) + 1-s
	ys = coordSum(
         coordQuotient(const_times_coord(xm, coordProduct(c_s, c_t)), denom),
         const_times_coord(x1/(xm-x1),coordProduct(oms, c_t))
         ); // ys = xm*s*t/denom + (1-s)*t*x1/(xm-x1)
	
	xys_To_xy(xm, xs, ys, x, y);
}
//---------------------------------------------------------------
void Dcoord_1(parameters par, parameters Dpar, int i, int j, double s, double t,
              derivs_1D G, derivs_1D DG, coordinate *x, coordinate *y, coordinate *Dx, coordinate *Dy)
{// Linear version of 'coord_1'
	double x1=X1FACT*par.xi,  x0=X0FACT*par.xo, xm=Func_xm(par), Dxm=Func_Dxm(par, Dpar);
	coordinate c_s, c_t, oms, omt, denom, Ddenom, xs, ys, Dxs, Dys;

	c_s.d0 = s; c_s.d1 = 1.; c_s.d2 = c_s.d11 = c_s.d12 = c_s.d22 = 0.;
	c_t.d0 = t; c_t.d2 = 1.; c_t.d1 = c_t.d11 = c_t.d12 = c_t.d22 = 0.;
	oms.d0 = 1.-s; oms.d1 = -1.; oms.d2 = oms.d11 = oms.d12 = oms.d22 = 0.;
	omt.d0 = 1.-t; omt.d2 = -1.; omt.d1 = omt.d11 = omt.d12 = omt.d22 = 0.;

	denom.d0  = x0 + xm*(1.-2.*t); // denom = x0 + xm*(1-2t)
	denom.d2  = -2.*xm;
	denom.d1  = denom.d11 = denom.d12 = denom.d22 = 0.;

	Ddenom.d0  = Dxm*(1.-2.*t);
	Ddenom.d2  = -2.*Dxm;
	Ddenom.d1  = Ddenom.d11 = Ddenom.d12 = Ddenom.d22 = 0.;

	xs  = coordSum(coordQuotient(const_times_coord(xm, coordProduct(c_s, omt)), denom) , oms); // xs=xm*s*(1-t)/(denom) + 1-s
	Dxs = DcoordQuotient(const_times_coord(xm, coordProduct(c_s, omt)), denom,
                        const_times_coord(Dxm, coordProduct(c_s, omt)), Ddenom);

	ys = coordSum(
         coordQuotient(const_times_coord(xm, coordProduct(c_s, c_t)), denom),
         const_times_coord(x1/(xm-x1),coordProduct(oms, c_t))
         ); // ys = xm*s*t/denom + (1-s)*t*x1/(xm-x1)

	Dys = coordSum(
         DcoordQuotient(const_times_coord(xm, coordProduct(c_s, c_t)), denom,
                        const_times_coord(Dxm, coordProduct(c_s, c_t)), Ddenom),
         const_times_coord(-Dxm*x1/(xm-x1)/(xm-x1),coordProduct(oms, c_t))
         );

	
	D_xys_To_xy(xm, Dxm, xs, ys, Dxs, Dys, x, y, Dx, Dy);
}
//   -----------------------------------------------
void coord_2(parameters par, int i, int j, double s, double t, derivs_1D G, coordinate *x, coordinate *y)
{// Coordinate transformation in domain 3
	double xm=Func_xm(par), tm1=t-1.;
	coordinate xs, ys, x1, xs1, y1, ys1, x3, xs3, y3, ys3, a, b;

	coord_1(par, i, j, s, 1., G, &x1, &y1);
	xy_To_xys(xm, x1, y1, &xs1, &ys1);
	
	coord_3(par, i, j, s, 0., G, &x3, &y3);
	xy_To_xys(xm, x3, y3, &xs3, &ys3);

	// setting all t-derivatives to 0:
	xs1.d2=xs1.d22=xs1.d12=xs3.d2=xs3.d22=xs3.d12=0.;
	ys1.d2=ys1.d22=ys1.d12=ys3.d2=ys3.d22=ys3.d12=0.;

	xs.d0  = t*xs1.d0  - tm1*xs3.d0;           // xs = t*xs1 + (1-t)*xs3
	xs.d1  = t*xs1.d1  - tm1*xs3.d1;
	xs.d11 = t*xs1.d11 - tm1*xs3.d11;
	xs.d2  =   xs1.d0  - xs3.d0;
	xs.d12 =   xs1.d1  - xs3.d1;
	xs.d22 = 0.;

	a      = coordLog(coordQuotient(ys1,ys3)); // a = Log(ys1/ys3)
	b.d0   = t*a.d0;                           // b = t*Log(ys1/ys3)
	b.d1   = t*a.d1;
	b.d2   = a.d0;
	b.d11  = t*a.d11;
	b.d22  = 0.;
	b.d12  = a.d1;
	ys     = coordProduct(ys3, coordExp(b));  // ys = ys3*(ys1/ys3)^t

	xys_To_xy(xm, xs, ys, x, y);
}
//---------------------------------------------------------------
void Dcoord_2(parameters par, parameters Dpar, int i, int j, double s, double t,
              derivs_1D G, derivs_1D DG, coordinate *x, coordinate *y, coordinate *Dx, coordinate *Dy)
{// Linear version of 'coord_3'
	double xm=Func_xm(par), Dxm=Func_Dxm(par, Dpar), tm1=t-1.;
	coordinate   xs,  ys,  x1,  xs1,  y1,  ys1,  x3,  xs3,  y3,  ys3,  a0,  a,  b,  c,
				Dxs, Dys, Dx1, Dxs1, Dy1, Dys1, Dx3, Dxs3, Dy3, Dys3, Da0, Da, Db, Dc;

	Dcoord_1(par, Dpar, i, j, s, 1., G, DG, &x1, &y1, &Dx1, &Dy1);
	D_xy_To_xys(xm, Dxm, x1, y1, Dx1, Dy1, &xs1, &ys1, &Dxs1, &Dys1);
	
	Dcoord_3(par, Dpar, i, j, s, 0., G, DG, &x3, &y3, &Dx3, &Dy3);
	D_xy_To_xys(xm, Dxm, x3, y3, Dx3, Dy3, &xs3, &ys3, &Dxs3, &Dys3);
	
	// setting all t-derivatives to 0:
	xs1.d2 =xs1.d22 =xs1.d12 =xs3.d2 =xs3.d22 =xs3.d12 =0.;
	ys1.d2 =ys1.d22 =ys1.d12 =ys3.d2 =ys3.d22 =ys3.d12 =0.;
	Dxs1.d2=Dxs1.d22=Dxs1.d12=Dxs3.d2=Dxs3.d22=Dxs3.d12=0.;
	Dys1.d2=Dys1.d22=Dys1.d12=Dys3.d2=Dys3.d22=Dys3.d12=0.;

	// xs = t*xs1 + (1-t)*xs3:
	xs.d0  = t*xs1.d0  - tm1*xs3.d0;    Dxs.d0  = t*Dxs1.d0  - tm1*Dxs3.d0;
	xs.d1  = t*xs1.d1  - tm1*xs3.d1;    Dxs.d1  = t*Dxs1.d1  - tm1*Dxs3.d1;
	xs.d11 = t*xs1.d11 - tm1*xs3.d11;   Dxs.d11 = t*Dxs1.d11 - tm1*Dxs3.d11;
	xs.d2  =   xs1.d0  - xs3.d0;        Dxs.d2  =   Dxs1.d0  - Dxs3.d0;
	xs.d12 =   xs1.d1  - xs3.d1;        Dxs.d12 =   Dxs1.d1  - Dxs3.d1;
	xs.d22 = 0.;                        Dxs.d22 = 0.;

	// ys = ys3*(ys1/ys3)^t  :
	a0     = coordQuotient(ys1,ys3);    Da0     = DcoordQuotient(ys1,ys3,Dys1,Dys3);
	a      = coordLog(a0);              Da      = DcoordLog(a0, Da0); // a = Log(ys1/ys3)
	b.d0   = t*a.d0;                    Db.d0   = t*Da.d0;            // b = t*Log(ys1/ys3)
	b.d1   = t*a.d1;                    Db.d1   = t*Da.d1;
	b.d2   = a.d0;                      Db.d2   = Da.d0;
	b.d11  = t*a.d11;                   Db.d11  = t*Da.d11;
	b.d22  = 0.;                        Db.d22  = 0.;
	b.d12  = a.d1;                      Db.d12  = Da.d1;
	c      = coordExp(b);               Dc      = DcoordExp(b, Db);   // c = Exp(b)
	ys     = coordProduct(ys3, c);      Dys     = DcoordProduct(ys3, c, Dys3, Dc);

	D_xys_To_xy(xm, Dxm, xs, ys, Dxs, Dys, x, y, Dx, Dy);
}
//   -----------------------------------------------
void coord_2_all_st(parameters par, int i, int j, double s, double t, derivs_1D Gcheb, coordinate *x, coordinate *y)
{// Coordinate transformation in domain 3
	double xm=Func_xm(par), tm1=t-1.;
	coordinate xs, ys, x1, xs1, y1, ys1, x3, xs3, y3, ys3, a, b;

	coord_1(par, i, j, s, 1., Gcheb, &x1, &y1); // Gcheb plays no role here
	xy_To_xys(xm, x1, y1, &xs1, &ys1);
	if (s<TINY) xs1.d0 = 1.;
	
	coord_3_all_st(par, s, 0., Gcheb, &x3, &y3);
	xy_To_xys(xm, x3, y3, &xs3, &ys3);
//	if (s<TINY) xs3.d0 = par.xo;
	if (s<TINY) xs3.d0 = 1.;

	// setting all t-derivatives to 0:
	xs1.d2=xs1.d22=xs1.d12=xs3.d2=xs3.d22=xs3.d12=0.;
	ys1.d2=ys1.d22=ys1.d12=ys3.d2=ys3.d22=ys3.d12=0.;

	xs.d0  = t*xs1.d0  - tm1*xs3.d0;           // xs = t*xs1 + (1-t)*xs3
	xs.d1  = t*xs1.d1  - tm1*xs3.d1;
	xs.d11 = t*xs1.d11 - tm1*xs3.d11;
	xs.d2  =   xs1.d0  - xs3.d0;
	xs.d12 =   xs1.d1  - xs3.d1;
	xs.d22 = 0.;

	a      = coordLog(coordQuotient(ys1,ys3)); // a = Log(ys1/ys3)
	b.d0   = t*a.d0;                           // b = t*Log(ys1/ys3)
	b.d1   = t*a.d1;
	b.d2   = a.d0;
	b.d11  = t*a.d11;
	b.d22  = 0.;
	b.d12  = a.d1;
	ys     = coordProduct(ys3, coordExp(b));  // ys = ys3*(ys1/ys3)^t

	xys_To_xy(xm, xs, ys, x, y);
}
//   -----------------------------------------------
void coord_3(parameters par, int i, int j, double s, double t, derivs_1D G, coordinate *x, coordinate *y)
{// Coordinate transformation in domain 3
	double tm1=t-1.,
		xi     = par.xi,
		xo     = par.xo,
		mant   = log(xo/xi),
		sig    = exp(s*mant),
		sig_s  = mant*sig,
		sig_ss = mant*sig_s;

	(*x).d0   = xi*sig;
	(*x).d1   = xi*sig_s;
	(*x).d11  = xi*sig_ss;
	(*x).d2   = (*x).d12 = (*x).d22 = 0.;
	
	(*y).d22  = 0.;
	(*y).d2   = -(G.d0[i]  - (*x).d0);
	(*y).d12  = -(G.d1[i]  - (*x).d1);
	(*y).d112 = -(G.d11[i] - (*x).d11);
	(*y).d0   = tm1*(*y).d2;
	(*y).d1   = tm1*(*y).d12;
	(*y).d11  = tm1*(*y).d112;
}
//---------------------------------------------------------------
void Dcoord_3(parameters par, parameters Dpar, int i, int j, double s, double t,
              derivs_1D G, derivs_1D DG, coordinate *x, coordinate *y, coordinate *Dx, coordinate *Dy)
{// Linear version of 'coord_3'
	double tm1=t-1.,
		xi     = par.xi,                    Dxi        = Dpar.xi,
		xo     = par.xo,                    Dxo        = Dpar.xo,
		mant   = log(xo/xi),                Dmant      = Dxo/xo-Dxi/xi,
		sig    = exp(mant*s),               Dsig       = sig*s*Dmant,
		sig_s  = mant*sig,                  Dsig_s     = Dmant*sig   + mant*Dsig,
		sig_ss = mant*sig_s,                Dsig_ss    = Dmant*sig_s + mant*Dsig_s;

	(*x).d0   = xi*sig;                     (*Dx).d0  = Dxi*sig    + xi*Dsig;
	(*x).d1   = xi*sig_s;                   (*Dx).d1  = Dxi*sig_s  + xi*Dsig_s;
	(*x).d11  = xi*sig_ss;                  (*Dx).d11 = Dxi*sig_ss + xi*Dsig_ss;
	(*x).d2   = (*x).d12 = (*x).d22 = 0;    (*Dx).d2  = (*Dx).d12 = (*Dx).d22 = 0.;
	
	(*y).d22  = 0.;                        (*Dy).d22  = 0.;
	(*y).d2   = -(G.d0[i]  - (*x).d0);     (*Dy).d2   = -(DG.d0[i]  - (*Dx).d0);
	(*y).d12  = -(G.d1[i]  - (*x).d1);     (*Dy).d12  = -(DG.d1[i]  - (*Dx).d1);
	(*y).d112 = -(G.d11[i] - (*x).d11);    (*Dy).d112 = -(DG.d11[i] - (*Dx).d11);
	(*y).d0   = tm1*(*y).d2;               (*Dy).d0   = tm1*(*Dy).d2;
	(*y).d1   = tm1*(*y).d12;              (*Dy).d1   = tm1*(*Dy).d12;
	(*y).d11  = tm1*(*y).d112;             (*Dy).d11  = tm1*(*Dy).d112;
}
//  -----------------------------------------------
void coord_3_all_st(parameters par, double s, double t, derivs_1D Gcheb, coordinate *x, coordinate *y)
{// Coordinate transformation in domain 3
	double tm1=t-1., G_val, G1_val, G11_val,
		xi     = par.xi,
		xo     = par.xo,
		mant   = log(xo/xi),
		sig    = exp(s*mant),
		sig_s  = mant*sig,
		sig_ss = mant*sig_s;

	(*x).d0   = xi*sig;
	(*x).d1   = xi*sig_s;
	(*x).d11  = xi*sig_ss;
	(*x).d2   = (*x).d12 = (*x).d22 = 0.;
	

/*	chebft_Extremes(G.d0, par.ns[NDOM-1], 0); // create Chebyshev coefficients for G
	chebft_Extremes(G.d1, par.ns[NDOM-1], 0);
	chebft_Extremes(G.d11, par.ns[NDOM-1], 0);*/
  G_val = chebev(0., 1., Gcheb.d0, par.ns[NDOM-1], s);
  G1_val = chebev(0., 1., Gcheb.d1, par.ns[NDOM-1], s);
  G11_val = chebev(0., 1., Gcheb.d11, par.ns[NDOM-1], s);
// 	chebft_Extremes(G.d0, par.ns[NDOM-1], 1); // create functional values again
// 	chebft_Extremes(G.d1, par.ns[NDOM-1], 1);
// 	chebft_Extremes(G.d11, par.ns[NDOM-1], 1);

	(*y).d22  = 0.;
	(*y).d2   = -(G_val  - (*x).d0);
	(*y).d12  = -(G1_val  - (*x).d1);
	(*y).d112 = -(G11_val - (*x).d11);
	(*y).d0   = tm1*(*y).d2;
	(*y).d1   = tm1*(*y).d12;
	(*y).d11  = tm1*(*y).d112;
}
//   -----------------------------------------------
void coord_3_all_st_old(parameters par, double s, double t, derivs_1D G, coordinate *x, coordinate *y)
{// Coordinate transformation in domain 3
	double tm1=t-1., G_val, G1_val, G11_val,
		xi     = par.xi,
		xo     = par.xo,
		mant   = log(xo/xi),
		sig    = exp(s*mant),
		sig_s  = mant*sig,
		sig_ss = mant*sig_s;

	(*x).d0   = xi*sig;
	(*x).d1   = xi*sig_s;
	(*x).d11  = xi*sig_ss;
	(*x).d2   = (*x).d12 = (*x).d22 = 0.;
	

	chebft_Extremes(G.d0, par.ns[NDOM-1], 0); // create Chebyshev coefficients for G
	chebft_Extremes(G.d1, par.ns[NDOM-1], 0);
	chebft_Extremes(G.d11, par.ns[NDOM-1], 0);
  G_val = chebev(0., 1., G.d0, par.ns[NDOM-1], s);
  G1_val = chebev(0., 1., G.d1, par.ns[NDOM-1], s);
  G11_val = chebev(0., 1., G.d11, par.ns[NDOM-1], s);
	chebft_Extremes(G.d0, par.ns[NDOM-1], 1); // create functional values again
	chebft_Extremes(G.d1, par.ns[NDOM-1], 1);
	chebft_Extremes(G.d11, par.ns[NDOM-1], 1);

	(*y).d22  = 0.;
	(*y).d2   = -(G_val  - (*x).d0);
	(*y).d12  = -(G1_val  - (*x).d1);
	(*y).d112 = -(G11_val - (*x).d11);
	(*y).d0   = tm1*(*y).d2;
	(*y).d1   = tm1*(*y).d12;
	(*y).d11  = tm1*(*y).d112;
}
//  -----------------------------------------------
st coord_xy_To_st(parameters par, derivs_1D G, derivs_1D Gcheb, double x, double y)
{// Calculation of the coordinates (s,t) from (x,y)
	st st_pair;
	int solved=0, dummy_int;
	double xi = par.xi, xo = par.xo, x0=X0FACT*par.xo, x1=X1FACT*par.xi,
         xm=Func_xm(par), xs, ys, s, s0, s1, smin=0., smax=1., t, t0, t1, denom, G_val,
				 xs1, ys1, xs3, ys3, test, test_old=1., tolerance=1.e-9;
	coordinate x_temp, y_temp, xs_temp, ys_temp;
	derivs_1D dummy_deriv;

	st_pair.idom=-12; // incorrect starting value to make sure it's then corrected

 // calculation of x-tilde and y-tilde from x and y
	denom = sqrt((xm-x)*(xm-x) + 2.*y*(xm+x) + y*y);
	xs = 0.5*(1.+(xm-x-y)/denom);
	ys = 0.5*(-1.+(xm+x+y)/denom);

 // first I determine if x+y>=x0 and therefore in domain 0
	if (x+y>=x0){
		st_pair.idom=0;
		s = sqrt(x0/(x+y));
		t = ys*(x0+xm*s*s)/(xm*s*s*(1.+2.*ys));
		if (s>=-tolerance && s<=1.+tolerance && t>=-tolerance && t<=1.+tolerance) solved=1; else printf("Problem in domain 0\n");
	}

 // next I check to see if the point might be in the ring
	if (solved==0 && x>=xi && x<=xo){
		s = log(x/xi)/log(xo/xi); // this s corresponds to the correct value of x
//		chebft_Extremes(G.d0, par.ns[NDOM-1], 0); // create Chebyshev coefficients for G
		G_val = chebev(0., 1., Gcheb.d0, par.ns[NDOM-1], s);
		if (fabs(x-G_val)<tolerance){
			if (y<tolerance) {t=0.; st_pair.idom=NDOM-2; solved=1;} // one of the points near xi or xo on the surface
			else t=12.; // the point is outside the ring
		}
		else t = 1.-y/(G_val-x);
		if (s>=-tolerance && s<=1.+tolerance && t>=-tolerance && t<=1.+tolerance && solved==0) 
			{solved=1; st_pair.idom=NDOM-1;}
//		chebft_Extremes(G.d0, par.ns[NDOM-1], 1); // create functional values again for G
	}

// next I test if it is in domain 1
	if (solved==0){
		if (x<=x1) st_pair.idom = 1; // we're definitely in this domain
		// the intersection of the correct radial distance with the boundary surface is determined
		else{
			s = (x0-xm)*(x1-x-y)/((x0-xm)*(x1-x-y) + (xm-x1)*(x+y-x0));
			coord_1(par, dummy_int, dummy_int, s, 1.0, dummy_deriv, &x_temp, &y_temp);
			if (y_temp.d0 < y) st_pair.idom = 1;
		}
		if (st_pair.idom == 1){
			// we're in domain 1, now s is determined for the correct radial distance and using bisection, t is found
			t0=0.; t1=1.; t=0.5*(t0+t1);
			while (fabs(t0-t1)>1.e-15){
				s = (2.*xm*t-x0-xm)*((x+y)*(xm-x1*(1.-t))-x1*xm*t)/((x+y-xm)*(2.*t*t*x1*xm-x1*t*(xm+x0)-x0*(x1-xm))+
								2.*(x+y)*(xm-x1)*(xm*t-x0));
				coord_1(par, dummy_int, dummy_int, s, t, dummy_deriv, &x_temp, &y_temp);
				if (x_temp.d0 < x) t0 = t;
				else            t1 = t;
				t = 0.5*(t0+t1);
			}
			if (s>=-tolerance && s<=1.+tolerance && t>=-tolerance && t<=1.+tolerance) solved=1;
			else printf("Problem in domain 1: s=%e t=%e\n",s,t);
		}
	}

	if (solved==0){ // the point must be in domain 2
		if (y<tolerance){
			if (x<=xi) s=0.;
			else       s=1.;
			coord_1(par, dummy_int, dummy_int, s, 1., Gcheb, &x_temp, &y_temp);
			xy_To_xys(xm, x_temp, y_temp, &xs_temp, &ys_temp);
			ys1 = ys_temp.d0;
			if (fabs(x-xi)<tolerance || fabs(x-xo)<tolerance) t=0.;
			else{
				if (x<=xi) coord_3(par, 0 , 0, 0., 0., G, &x_temp, &y_temp);
				else       coord_3(par, par.ns[NDOM-1]-1 , 0, 1., 0., G, &x_temp, &y_temp);
				xy_To_xys(xm, x_temp, y_temp, &xs_temp, &ys_temp);
				ys3 = ys_temp.d0;
				t = log(ys/ys3)/log(ys1/ys3);
			}
		}
		else{
			smax = 1.-xs; // t=(xs-xs3)/(xs1-xs3)<=1  --> s<= 1-xs if numerator and denominator
                    // positive, see below
			// now smin is determined by solving the equation xs=xs3
			coord_1(par, dummy_int, dummy_int, smax, 1., G, &x_temp, &y_temp);
			xy_To_xys(xm, x_temp, y_temp, &xs_temp, &ys_temp);
			xs1 = xs_temp.d0;
			ys1 = ys_temp.d0;
			coord_3_all_st(par, smax, 0., Gcheb, &x_temp, &y_temp);
			xy_To_xys(xm, x_temp, y_temp, &xs_temp, &ys_temp);
			xs3 = xs_temp.d0;
			ys3 = ys_temp.d0;
			test_old = log(ys1)*(xs-xs3)-log(ys3)*xs - log(ys)*(xs1-xs3) + log(ys3)*xs1;
			if (xs1-xs3<0.){smin=smax; smax=1.; test_old=-test_old;} //
			s0=smin; s1=smax; s=0.5*(s0+s1);
			while (fabs(s0-s1)>1.e-15){
				coord_1(par, dummy_int, dummy_int, s, 1., G, &x_temp, &y_temp);
				xy_To_xys(xm, x_temp, y_temp, &xs_temp, &ys_temp);
				xs1 = xs_temp.d0;
				ys1 = ys_temp.d0;
				coord_3_all_st(par, s, 0., Gcheb, &x_temp, &y_temp);
				xy_To_xys(xm, x_temp, y_temp, &xs_temp, &ys_temp);
				xs3 = xs_temp.d0;
				ys3 = ys_temp.d0;
				test = log(ys1)*(xs-xs3)-log(ys3)*xs - log(ys)*(xs1-xs3) + log(ys3)*xs1;

				if (test*test_old >0.) s1=s;
				else         s0=s;

				s = 0.5*(s0+s1);
			}
			t = log(ys/ys3)/log(ys1/ys3);
		}
		if (s>=-tolerance && s<=1.+tolerance && t>=-tolerance && t<=1.+tolerance)
       {solved=1; st_pair.idom=2;}
		else printf("Problem in domain 2: x=%e y=%e s=%e t=%e\n",x,y,s,t);
	}

	if (solved==1){
		st_pair.s=s; st_pair.t=t;
	// test of solution
/*		if (st_pair.idom==0) coord_0(par, dummy_int, dummy_int, s, t, G, &x_temp, &y_temp);
		if (st_pair.idom==1) coord_1(par, dummy_int, dummy_int, s, t, G, &x_temp, &y_temp);
		if (st_pair.idom==2) coord_2_all_st(par, dummy_int, dummy_int, s, t, G, &x_temp, &y_temp);
		if (st_pair.idom==3) coord_3_all_st(par, s, t, Gcheb, &x_temp, &y_temp);
		if (fabs(x_temp.d0-x)>1.e-8 || fabs(y_temp.d0-y)>1.e-8){
			printf("delx=%12.12e \t dely=%12.12e\n",x-x_temp.d0, y-y_temp.d0);
			if (st_pair.idom==2) printf("dom=2: smin=%e smax=%e s=%e t=%e\n",smin,smax,s,t);
		}*/
	}
	else printf("The routine coord_xy_To_st failed to find an (s,t) pair corresponding to x=%e y=%e\n",x,y);
	return st_pair;
}
//  -----------------------------------------------
//  -----------------------------------------------
st coord_xy_To_st_old(parameters par, derivs_1D G, double x, double y)
{// Calculation of the coordinates (s,t) from (x,y)
	st st_pair;
	int solved=0, dummy_int;
	double xi = par.xi, xo = par.xo, x0=X0FACT*par.xo, x1=X1FACT*par.xi,
         xm=Func_xm(par), xs, ys, s, s0, s1, smin=0., smax=1., t, t0, t1, denom, G_val,
				 xs1, ys1, xs3, ys3, test, test_old=1., tolerance=1.e-9;
	coordinate x_temp, y_temp, xs_temp, ys_temp;
	derivs_1D dummy_deriv;

	st_pair.idom=-12; // incorrect starting value to make sure it's then corrected

 // calculation of x-tilde and y-tilde from x and y
	denom = sqrt((xm-x)*(xm-x) + 2.*y*(xm+x) + y*y);
	xs = 0.5*(1.+(xm-x-y)/denom);
	ys = 0.5*(-1.+(xm+x+y)/denom);

 // first I determine if x+y>=x0 and therefore in domain 0
	if (x+y>=x0){
		st_pair.idom=0;
		s = sqrt(x0/(x+y));
		t = ys*(x0+xm*s*s)/(xm*s*s*(1.+2.*ys));
		if (s>=-tolerance && s<=1.+tolerance && t>=-tolerance && t<=1.+tolerance) solved=1; else printf("Problem in domain 0\n");
	}

 // next I check to see if the point might be in the ring
	if (solved==0 && x>=xi && x<=xo){
		s = log(x/xi)/log(xo/xi); // this s corresponds to the correct value of x
		chebft_Extremes(G.d0, par.ns[NDOM-1], 0); // create Chebyshev coefficients for G
		G_val = chebev(0., 1., G.d0, par.ns[NDOM-1], s);
		if (fabs(x-G_val)<1.e-10) t=0.; // one of the points near xi or xo on the surface
		else t = 1.-y/(G_val-x);
		if (s>=-tolerance && s<=1.+tolerance && t>=-tolerance && t<=1.+tolerance) 
			{solved=1; st_pair.idom=NDOM-1;}
		chebft_Extremes(G.d0, par.ns[NDOM-1], 1); // create functional values again for G
	}

// next I test if it is in domain 1
	if (solved==0){
		if (x<=x1) st_pair.idom = 1; // we're definitely in this domain
		// the intersection of the correct radial distance with the boundary surface is determined
		else{
			s = (x0-xm)*(x1-x-y)/((x0-xm)*(x1-x-y) + (xm-x1)*(x+y-x0));
			coord_1(par, dummy_int, dummy_int, s, 1.0, dummy_deriv, &x_temp, &y_temp);
			if (y_temp.d0 < y) st_pair.idom = 1;
		}
		if (st_pair.idom == 1){
			// we're in domain 1, now s is determined for the correct radial distance and using bisection, t is found
			t0=0.; t1=1.; t=0.5*(t0+t1);
			while (fabs(t0-t1)>1.e-15){
				s = (2.*xm*t-x0-xm)*((x+y)*(xm-x1*(1.-t))-x1*xm*t)/((x+y-xm)*(2.*t*t*x1*xm-x1*t*(xm+x0)-x0*(x1-xm))+
								2.*(x+y)*(xm-x1)*(xm*t-x0));
				coord_1(par, dummy_int, dummy_int, s, t, dummy_deriv, &x_temp, &y_temp);
				if (x_temp.d0 < x) t0 = t;
				else            t1 = t;
				t = 0.5*(t0+t1);
			}
			if (s>=-tolerance && s<=1.+tolerance && t>=-tolerance && t<=1.+tolerance) solved=1;
			else printf("Problem in domain 1: s=%e t=%e\n",s,t);
		}
	}

	if (solved==0){ // the point must be in domain 2
		if (y<tolerance){
			if (x<=xi) s=0.;
			else       s=1.;
			coord_1(par, dummy_int, dummy_int, s, 1., G, &x_temp, &y_temp);
			xy_To_xys(xm, x_temp, y_temp, &xs_temp, &ys_temp);
			ys1 = ys_temp.d0;
			coord_3_all_st(par, s, 0., G, &x_temp, &y_temp);
			xy_To_xys(xm, x_temp, y_temp, &xs_temp, &ys_temp);
			ys3 = ys_temp.d0;
			t = log(ys/ys3)/log(ys1/ys3);
		}
		else{
			smax = 1.-xs; // t=(xs-xs3)/(xs1-xs3)<=1  --> s<= 1-xs if numerator and denominator
                    // positive, see below
			// now smin is determined by solving the equation xs=xs3
			coord_1(par, dummy_int, dummy_int, smax, 1., G, &x_temp, &y_temp);
			xy_To_xys(xm, x_temp, y_temp, &xs_temp, &ys_temp);
			xs1 = xs_temp.d0;
			ys1 = ys_temp.d0;
			coord_3_all_st(par, smax, 0., G, &x_temp, &y_temp);
			xy_To_xys(xm, x_temp, y_temp, &xs_temp, &ys_temp);
			xs3 = xs_temp.d0;
			ys3 = ys_temp.d0;
			test_old = log(ys1)*(xs-xs3)-log(ys3)*xs - log(ys)*(xs1-xs3) + log(ys3)*xs1;
			if (xs1-xs3<0.){smin=smax; smax=1.; test_old=-test_old;} //
			s0=smin; s1=smax; s=0.5*(s0+s1);
			while (fabs(s0-s1)>1.e-15){
				coord_1(par, dummy_int, dummy_int, s, 1., G, &x_temp, &y_temp);
				xy_To_xys(xm, x_temp, y_temp, &xs_temp, &ys_temp);
				xs1 = xs_temp.d0;
				ys1 = ys_temp.d0;
				coord_3_all_st(par, s, 0., G, &x_temp, &y_temp);
				xy_To_xys(xm, x_temp, y_temp, &xs_temp, &ys_temp);
				xs3 = xs_temp.d0;
				ys3 = ys_temp.d0;
				test = log(ys1)*(xs-xs3)-log(ys3)*xs - log(ys)*(xs1-xs3) + log(ys3)*xs1;

				if (test*test_old >0.) s1=s;
				else         s0=s;

				s = 0.5*(s0+s1);
			}
			t = log(ys/ys3)/log(ys1/ys3);
		}
		if (s>=-tolerance && s<=1.+tolerance && t>=-tolerance && t<=1.+tolerance)
       {solved=1; st_pair.idom=2;}
		else printf("Problem in domain 2: x=%e y=%e s=%e t=%e\n",x,y,s,t);
	}

	if (solved==1){
		st_pair.s=s; st_pair.t=t;
	// test of solution
/*		if (st_pair.idom==0) coord_0(par, dummy_int, dummy_int, s, t, G, &x_temp, &y_temp);
		if (st_pair.idom==1) coord_1(par, dummy_int, dummy_int, s, t, G, &x_temp, &y_temp);
		if (st_pair.idom==2) coord_2_all_st(par, dummy_int, dummy_int, s, t, G, &x_temp, &y_temp);
		if (st_pair.idom==3) coord_3_all_st(par, s, t, G, &x_temp, &y_temp);
		if (fabs(x_temp.d0-x)>1.e-8 || fabs(y_temp.d0-y)>1.e-8){
			printf("delx=%12.12e \t dely=%12.12e\n",x-x_temp.d0, y-y_temp.d0);
			if (st_pair.idom==2) printf("dom=2: smin=%e smax=%e s=%e t=%e\n",smin,smax,s,t);
		}*/
	}
	else printf("The routine coord_xy_To_st failed to find an (s,t) pair corresponding to x=%e y=%e\n",x,y);
	return st_pair;
}
//  -----------------------------------------------
//  ***** Auxiliary coordinate manipulations ******
//  -----------------------------------------------

//  -----------------------------------------------
coordinate coordSum(coordinate a, coordinate b)
{
	coordinate result;
	
	result.d0   = a.d0    + b.d0;
	result.d1   = a.d1    + b.d1;
	result.d2   = a.d2    + b.d2;
	result.d11  = a.d11   + b.d11;
	result.d22  = a.d22   + b.d22;
	result.d12  = a.d12   + b.d12;
	result.d112 = 0.; // not needed

	return result;
}
//  -----------------------------------------------
coordinate coordDiff(coordinate a, coordinate b)
{
	coordinate result;
	
	result.d0   = a.d0   - b.d0;
	result.d1   = a.d1   - b.d1;
	result.d2   = a.d2   - b.d2;
	result.d11  = a.d11  - b.d11;
	result.d22  = a.d22  - b.d22;
	result.d12  = a.d12  - b.d12;
	result.d112 = 0.; // not needed

	return result;
}
//  -----------------------------------------------
coordinate coordInverse(coordinate a)
{
	double 
		aux1 = 1./a.d0, 
		aux2 = aux1*aux1, 
		aux3 = aux1*aux2;
	coordinate result;
	
	result.d0   = aux1;
	result.d1   = -aux2*a.d1;
	result.d2   = -aux2*a.d2;
	result.d11  = 2.*aux3*a.d1*a.d1-aux2*a.d11;
	result.d22  = 2.*aux3*a.d2*a.d2-aux2*a.d22;
	result.d12  = 2.*aux3*a.d1*a.d2-aux2*a.d12;
	result.d112 = 0.; // not needed

	return result;
}
//  -----------------------------------------------
coordinate coordProduct(coordinate a, coordinate b)
{
	coordinate result;
	
	result.d0   = a.d0*b.d0;
	result.d1   = a.d1*b.d0+a.d0*b.d1;
	result.d2   = a.d2*b.d0+a.d0*b.d2;
	result.d11  = a.d11*b.d0+2.*a.d1*b.d1+a.d0*b.d11;
	result.d22  = a.d22*b.d0+2.*a.d2*b.d2+a.d0*b.d22;
	result.d12  = a.d12*b.d0+a.d1*b.d2+a.d2*b.d1+a.d0*b.d12;
	result.d112 = 0.; // not needed

	return result;
}
//  -----------------------------------------------
coordinate coordQuotient(coordinate a, coordinate b)
{
	return coordProduct(a, coordInverse(b));
}
//  -----------------------------------------------
coordinate coordRoot(coordinate a)
{
	double b=sqrt(a.d0), b1=1./b, b2=b1*b1;
	coordinate result;
	
	result.d0   = b;
	result.d1   = 0.5*a.d1*b1;
	result.d2   = 0.5*a.d2*b1;
	result.d11  = 0.5*(a.d11*b1-a.d1*result.d1*b2);
	result.d22  = 0.5*(a.d22*b1-a.d2*result.d2*b2);
	result.d12  = 0.5*(a.d12*b1-a.d1*result.d2*b2);
	result.d112 = 0.; // not needed

	return result;
}
//  -----------------------------------------------
coordinate coordLog(coordinate a)
{
	double b=1./a.d0;
	coordinate result;
	
	result.d0   = -log(b);
	result.d1   = a.d1*b;
	result.d2   = a.d2*b;
	result.d11  = a.d11*b - result.d1*result.d1;
	result.d22  = a.d22*b - result.d2*result.d2;
	result.d12  = a.d12*b - result.d1*result.d2;
	result.d112 = 0.; // not needed

	return result;
}
//  -----------------------------------------------
coordinate coordExp(coordinate a)
{
	coordinate result;

	result.d0   = exp(a.d0);
	result.d1   = result.d0*a.d1;
	result.d2   = result.d0*a.d2;
	result.d11  = result.d1*a.d1 + result.d0*a.d11;
	result.d22  = result.d2*a.d2 + result.d0*a.d22;
	result.d12  = result.d2*a.d1 + result.d0*a.d12;
	result.d112 = 0.; // not needed

	return result;
}
//  -----------------------------------------------
coordinate const_times_coord(double b, coordinate a)
{
	coordinate result;
	
	result.d0   = b*a.d0;
	result.d1   = b*a.d1;
	result.d2   = b*a.d2;
	result.d11  = b*a.d11;
	result.d22  = b*a.d22;
	result.d12  = b*a.d12;
	result.d112 = 0.; // not needed

	return result;
}
//  -----------------------------------------------
coordinate const_plus_coord(double b, coordinate a)
{
	coordinate result=a;
	
	result.d0  += b;

	return result;
}
//  -----------------------------------------------
coordinate DcoordQuotient(coordinate a, coordinate b, coordinate Da, coordinate Db)
{
	coordinate 
		c    = coordInverse(b), 
		ac   = coordProduct(a, c), 
		acDb = coordProduct(ac, Db);
		
	return coordProduct(c, coordDiff(Da, acDb));
}
//  -----------------------------------------------
coordinate DcoordProduct(coordinate a, coordinate b, coordinate Da, coordinate Db)
{
	return coordSum(coordProduct(Da,b), coordProduct(Db,a));
}
//  -----------------------------------------------
coordinate DcoordRoot(coordinate a, coordinate Da)
{
	return const_times_coord(0.5, coordQuotient(Da, coordRoot(a)));
}
//  -----------------------------------------------
coordinate DcoordLog(coordinate a, coordinate Da)
{
	return coordQuotient(Da, a);
}
//  -----------------------------------------------
coordinate DcoordExp(coordinate a, coordinate Da)
{
	return coordProduct(Da, coordExp(a));
}
//  -----------------------------------------------
coordinate D_const_times_coord(double b, double Db, coordinate a, coordinate Da)
{
	return coordSum(const_times_coord(Db, a), const_times_coord(b, Da));
}
//  -----------------------------------------------
