#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <ctype.h>
#include <time.h>
#include "utilities.h"
#include "Ring.h"

//--------------------------------------------------------------------------------
int GradIndex(int i, int j)
{
	return 4*i+j;
}

// -------------------------------------------------------------------------------
void NonLinPotentials(parameters par, int idom, int i, int j, double s, double t,
	coordinate xx, coordinate yy, derivs_2D v, derivs_1D w, derivs_1D G, derivs_2D V, double *Laplace,
	double *grad_grad)
{	// Calculates the potential values, first and second (x,y)-derivatives as well as
	// the Laplacian of U at a given spatial grid point; for domain 4 see pages 12, 13
	double x = xx.d0, y = yy.d0;
	int ipot, m,n;
		
	for(ipot=0; ipot<NPOT; ipot ++){
		if (idom<NDOM-1){
			v_To_V(par, idom, i, j, v, V);
			pot_st_To_xy(par, xx, yy, V);
		}
		else{
			double Vs, Vss,
				xs = xx.d1,  xs2 = xs*xs,   xss  = xx.d11,
				ys = yy.d1,  ys2 = ys*ys,   yss  = yy.d11,
				yt = yy.d2,  yst = yy.d12,  ysst = yy.d112;
			int indx = Index(par,idom,ipot,i,j), indx_R = Index_Ring(par,ipot,i);
			V.d0[ipot]  = w.d0[indx_R]  + yt*v.d0[indx];
			Vs          = w.d1[indx_R]  + yst*v.d0[indx]  +     yt*v.d1[indx];
			Vss         = w.d11[indx_R] + ysst*v.d0[indx] + 2.*yst*v.d1[indx] + yt*v.d11[indx];
			V.d2[ipot]  = v.d2[indx];
			V.d1[ipot]  = (Vs-ys*V.d2[ipot])/xs;
			if (fabs(yt)>TINY) V.d22[ipot] = v.d22[indx]/yt; // V_yy not needed if yt=0
			V.d11[ipot] = (ys2*V.d22[ipot]+Vss-yss*V.d2[ipot]-2.*ys*v.d12[indx]-xss*V.d1[ipot])/xs2;
			V.d12[ipot] = -(ys*V.d22[ipot]-v.d12[indx])/xs;
		}
		Laplace[ipot]   = 4.*(x*V.d11[ipot]+y*V.d22[ipot]+V.d1[ipot]+0.5*V.d2[ipot]);
	}
	for (m=0; m<NPOT; m++){
		for (n=0; n<NPOT; n++){
			grad_grad[GradIndex(m,n)] = 4.*(x*V.d1[m]*V.d1[n]+y*V.d2[m]*V.d2[n]);
		}	
	}
}
// -------------------------------------------------------------------------------
void LinPotentials(parameters par, int idom, int i, int j, double s, double t,
 coordinate  xx,  coordinate yy, derivs_2D v,  derivs_1D w,  derivs_1D G,  derivs_2D V,  double *Laplace,  double *grad_grad,
 coordinate Dxx, coordinate Dyy, derivs_2D Dv, derivs_1D Dw, derivs_1D DG, derivs_2D DV, double *DLaplace, double *Dgrad_grad)
{	// Linear version of 'NonLinPotentials'
	double x = xx.d0, y = yy.d0, Dx = Dxx.d0, Dy = Dyy.d0;
	int ipot, m, n;
		
	for(ipot=0; ipot<NPOT; ipot ++){
		if (idom<NDOM-1){
			v_To_V(par, idom, i, j,  v,  V);
			v_To_V(par, idom, i, j, Dv, DV);
			D_pot_st_To_xy(par, xx, yy, V, Dxx, Dyy, DV);
		}
		else{
			double Vs, Vss, DVs, DVss,
				xs  = xx.d1,    xs2 = xs*xs,      xss = xx.d11,
				Dxs = Dxx.d1,                    Dxss = Dxx.d11,
				ys  = yy.d1,    ys2 = ys*ys,      yss = yy.d11,
				Dys = Dyy.d1,  Dys2 = 2.*ys*Dys, Dyss = Dyy.d11,
				yt  = yy.d2,    yst = yy.d12,    ysst = yy.d112,
				Dyt = Dyy.d2,  Dyst = Dyy.d12,  Dysst = Dyy.d112;

			int indx = Index(par,idom,ipot,i,j), indx_R = Index_Ring(par,ipot,i);

			V.d0[ipot]   = w.d0[indx_R]  +  yt*v.d0[indx];
			DV.d0[ipot]  = Dw.d0[indx_R] + Dyt*v.d0[indx]    + yt*Dv.d0[indx];
			Vs           = w.d1[indx_R]  + yst*v.d0[indx]    + yt*v.d1[indx];
			DVs          = Dw.d1[indx_R] + (Dyst*v.d0[indx]  + yst*Dv.d0[indx])
                                          + (Dyt*v.d1[indx] + yt*Dv.d1[indx]);
			Vss          = w.d11[indx_R] + ysst*v.d0[indx] + 2.*yst*v.d1[indx] + yt*v.d11[indx];
			DVss         = Dw.d11[indx_R] + (Dysst*v.d0[indx] + ysst*Dv.d0[indx])
                           + 2.*(Dyst*v.d1[indx] + yst*Dv.d1[indx])
                           + (Dyt*v.d11[indx] + yt*Dv.d11[indx]);
			V.d2[ipot]   = v.d2[indx];
			DV.d2[ipot]  = Dv.d2[indx];
			V.d1[ipot]   = (Vs-ys*V.d2[ipot])/xs;
			DV.d1[ipot]  = (DVs-Dys*V.d2[ipot]-ys*DV.d2[ipot]-Dxs*V.d1[ipot])/xs;
			if (fabs(yt)>TINY) {
				V.d22[ipot]  =   v.d22[indx]/yt;
				DV.d22[ipot] = (Dv.d22[indx] - Dyt*V.d22[ipot])/yt;
			}
			V.d11[ipot]  = ( ys2*V.d22[ipot]+Vss-yss*V.d2[ipot]-2.*ys*v.d12[indx]-xss*V.d1[ipot])/xs2;
			DV.d11[ipot] = (Dys2*V.d22[ipot]+ys2*DV.d22[ipot]+DVss-Dyss*V.d2[ipot]-yss*DV.d2[ipot]
				- 2.*Dys*v.d12[indx]-2.*ys*Dv.d12[indx]-Dxss*V.d1[ipot]-xss*DV.d1[ipot])/xs2 
				- 2.*V.d11[ipot]*Dxs/xs;
			V.d12[ipot]  = -(ys *V.d22[ipot]                 -v.d12[indx])/xs;
			DV.d12[ipot] = -(Dys*V.d22[ipot]+ys*DV.d22[ipot]-Dv.d12[indx])/xs - V.d12[ipot]*Dxs/xs;
		}
		Laplace[ipot]   = 4.*(x* V.d11[ipot]               +y* V.d22[ipot] 
								+  V.d1[ipot] + 0.5*V.d2[ipot]);
		DLaplace[ipot]  = 4.*(x*DV.d11[ipot]+Dx*V.d11[ipot]+y*DV.d22[ipot]+Dy*V.d22[ipot]
								+ DV.d1[ipot] + 0.5*DV.d2[ipot]);
	
	
	}
	for (m=0; m<NPOT; m++){
		for (n=0; n<NPOT; n++){
			grad_grad[GradIndex(m,n)]  = 4.*(x*V.d1[m]*V.d1[n]+y*V.d2[m]*V.d2[n]);
			Dgrad_grad[GradIndex(m,n)] = 4.*(Dx*V.d1[m]*V.d1[n]+x*DV.d1[m]*V.d1[n]+x*V.d1[m]*DV.d1[n]
			                               + Dy*V.d2[m]*V.d2[n]+y*DV.d2[m]*V.d2[n]+y*V.d2[m]*DV.d2[n]);
		}	
	}


}
// -------------------------------------------------------------------------------
void NonLinFieldEqns(parameters par, int idom, int i, int j,
		derivs_2D v, derivs_1D w, derivs_1D G, derivs_2D V, double *F)
{	// Nonlinear field equations and regularity of V_yy at s*(1-s)=0 in domain 4
	
	int ns=par.ns[idom], Ns=ns-1, nt=par.nt[idom], Nt=nt-1, ipot, indx;
//	p_eps p_e_pair;
	
	if (idom==NDOM-1 && i*(Ns-i) == 0 && j*(Nt-j) != 0){
		// Regularity condition of V_yy at s*(1-s)=0 in domain 4, see page 13
		// This condition is imposed if s*(1-s)=0 and t*(1-t) != 0
		for(ipot=0; ipot<NPOT; ipot ++){
			indx    = Index(par,idom,ipot,i,j);
			F[indx] = v.d22[indx]; // V_tt to be zeroed
		}
	}
	else{// Nonlinear field equations
//if (j==Nt && idom==4) j=0; 
		double Om=par.Om, V0=par.V0, Laplace[4], grad_grad[16], eps, epsB, p, h, x, y, u, ux, uy, B, Bx, By,
		       om, omx, omy, mu, mux, muy, vel, vel2, nu, e2V,
			a=Pih*i/Ns, sa=sin(a), s=sa*sa,
			b=Pih*j/Nt, sb=sin(b), t=sb*sb;
		coordinate xx, yy;
		coord_st_To_xy(par, idom, i, j, s, t, G, &xx, &yy);
		NonLinPotentials(par, idom, i, j, s, t, xx, yy, v, w, G, V, Laplace, grad_grad);
		x = xx.d0;    y = yy.d0;
		u = V.d0[0]; ux = V.d1[0]; uy = V.d2[0];
		B = V.d0[1]; Bx = V.d1[1]; By = V.d2[1];
		om= V.d0[2]; omx= V.d1[2]; omy= V.d2[2];
		mu= V.d0[3]; mux= V.d1[3]; muy= V.d2[3];

		// The equations [Bardeen, 1971, <Relativistic Disks>, p. 363]: 
		//    \nabla*(\nabla U)    = U_rho_rho   + U_zeta_zeta + U_rho/rho 
		//    (\nabla U)*(nabla V) = U_rho*V_rho + U_zeta*V_zeta

					
		//    (1) Equation (II.12)*e^(2*mu) with nu = u + ln(B)
		//    (2) Equation (II.13)*e^(2*mu)/W
		//    (3) Equation (II.14)  *e^(2*mu)
		
		F[Index(par,idom,0,i,j)] = Laplace[0] + (-2.*Bx + grad_grad[GradIndex(0,1)]
		                           -0.5*x*exp(-4.*u)/B*grad_grad[GradIndex(2,2)]) /B;

		F[Index(par,idom,1,i,j)] = Laplace[1] + 2.*Bx;
		
		F[Index(par,idom,2,i,j)] = Laplace[2] + 4.*omx - grad_grad[GradIndex(1,2)]/B - 4.*grad_grad[GradIndex(0,2)];
		
		F[Index(par,idom,3,i,j)] = Laplace[3] - 2.*mux - 2.*ux - 2.*Bx/B
				+ grad_grad[GradIndex(0,0)] + grad_grad[GradIndex(0,1)]/B - 0.25*x*exp(-4.*u)/B/B*grad_grad[GradIndex(2,2)];
											
		if (idom == NDOM-1){ // Matter terms if in ring
			nu  = u + log(B);
			vel = sqrt(x)*B*exp(-2.*nu)*(Om-om);
			vel2= vel*vel;
			e2V            = exp(2.*nu)*(1.-vel2);
			h           = par.V0 - 0.5*log(e2V);  // h    = V0 - V
			eos_quantities(par, h, &p, &eps, &epsB);// h(p) = Integrate[1/(eps(pp)+pp), {pp, 0, p}]
//			p_e_pair       = p_e(par,exp(V0)/sqrt(e2V));
//			p              = p_e_pair.p;
//			eps            = p_e_pair.eps;
			F[Index(par,idom,0,i,j)] -= 4.*Pi*exp(2.*mu)*((eps+p)*(1.+vel2)/(1.-vel2)-2.*p);
			F[Index(par,idom,1,i,j)] -= 16.*Pi*B*exp(2.*mu)*p;
			F[Index(par,idom,2,i,j)] -= -16.*Pi*B/sqrt(x)*exp(2.*mu+2.*u)*(eps+p)*vel/(1.-vel2);
			F[Index(par,idom,3,i,j)] -= -4.*Pi*exp(2.*mu)*(eps+p);
		}

// int k;
//if (idom==4){ 
// for (k=0; k<=3; k++){
// if (fabs(F[Index(par,idom,k,i,j)])>1.e-15){
// printf("NonLinFieldEqns: pot=%d \t dom=%d \t i=%d \t j=%d \t F=%e\n",k,idom,i,j,F[Index(par,idom,k,i,j)]);
// }
// }
//}

// if (idom==5)printf("NonLinFieldEqns: dom=%d \t i=%d \t j=%d \t Fu=%e \t FB=%e \t Fom=%e \t Fmu=%e\n",idom,i,j
// ,F[Index(par,idom,0,i,j)],F[Index(par,idom,1,i,j)],F[Index(par,idom,2,i,j)],F[Index(par,idom,3,i,j)]);

	}
}
// -------------------------------------------------------------------------------
void LinFieldEqns(parameters par, parameters Dpar, int idom, int i, int j, 
	derivs_2D  v, derivs_1D  w, derivs_1D  G, derivs_2D  V,
	derivs_2D Dv, derivs_1D Dw, derivs_1D DG, derivs_2D DV, double *JDX)
{	// Linear version of 'NonLinPotentials'
	int ns=par.ns[idom], Ns=ns-1, nt=par.nt[idom], Nt=nt-1, ipot, indx;
	double  Om=par.Om, DOm=Dpar.Om, V0=par.V0, DV0=Dpar.V0, DV_func,
			Laplace[4], grad_grad[16], eps, epsB, p, h, x,  y,  u,  ux,  uy,  B,  Bx,  By,  om,  omx,  omy,  mu,  mux,  muy,
	       DLaplace[4], Dgrad_grad[16], Deps, DepsB, Dp, Dh, Dx, Dy, Du, Dux, Duy, DB, DBx, DBy, Dom, Domx, Domy, Dmu, Dmux, Dmuy,
		   nu,   vel,  vel2,  e2V,
		   Dnu, Dvel, Dvel2, De2V,
		a=Pih*i/Ns, sa=sin(a), s=sa*sa,
		b=Pih*j/Nt, sb=sin(b), t=sb*sb;
	coordinate xx, yy, Dxx, Dyy;
//	p_eps p_e_pair, Dp_e_pair;
	
	if (idom==NDOM-1 && i*(Ns-i) == 0 && j*(Nt-j) != 0){
		for(ipot=0; ipot<NPOT; ipot ++){
			indx      = Index(par,idom,ipot,i,j);
			JDX[indx] = Dv.d22[indx];
		}
	}
	else{
		Dcoord_st_To_xy(par, Dpar, idom, i, j, s, t, G, DG, &xx, &yy, &Dxx, &Dyy);
		
		LinPotentials(par, idom, i, j, s, t, xx,  yy,  v,  w,  G,  V,  Laplace, grad_grad,
                                           Dxx, Dyy, Dv, Dw, DG, DV, DLaplace,  Dgrad_grad);
		x = xx.d0;    y = yy.d0;                 Dx = Dxx.d0;    Dy = Dyy.d0;
		u  = V.d0[0];  ux = V.d1[0];  uy = V.d2[0];  Du = DV.d0[0];  Dux = DV.d1[0];  Duy = DV.d2[0];
		B  = V.d0[1];  Bx = V.d1[1];  By = V.d2[1];  DB = DV.d0[1];  DBx = DV.d1[1];  DBy = DV.d2[1];
		om = V.d0[2]; omx = V.d1[2]; omy = V.d2[2]; Dom = DV.d0[2]; Domx = DV.d1[2]; Domy = DV.d2[2];
		mu = V.d0[3]; mux = V.d1[3]; muy = V.d2[3]; Dmu = DV.d0[3]; Dmux = DV.d1[3]; Dmuy = DV.d2[3];

		JDX[Index(par,idom,0,i,j)] = DLaplace[0] 
		 - DB*(-2.*Bx + grad_grad[GradIndex(0,1)] - 0.5*x*exp(-4.*u)/B*grad_grad[GradIndex(2,2)])/B/B
		 + (-2.*DBx + Dgrad_grad[GradIndex(0,1)] - 0.5*Dx*exp(-4.*u)/B*grad_grad[GradIndex(2,2)]
		      +  2.*Du*x*exp(-4.*u)/B*grad_grad[GradIndex(2,2)] + 0.5*DB*x*exp(-4.*u)/B/B*grad_grad[GradIndex(2,2)]
				- 0.5*x*exp(-4.*u)/B*Dgrad_grad[GradIndex(2,2)])/B;
		
		JDX[Index(par,idom,1,i,j)] = DLaplace[1] + 2.*DBx;
		
		JDX[Index(par,idom,2,i,j)] = DLaplace[2] + 4.*Domx - Dgrad_grad[GradIndex(1,2)]/B + DB*grad_grad[GradIndex(1,2)]/B/B
				- 4.*Dgrad_grad[GradIndex(0,2)];
		
		JDX[Index(par,idom,3,i,j)] = DLaplace[3] - 2.*Dmux - 2.*Dux - 2.*DBx/B + 2.*DB*Bx/B/B
				+ Dgrad_grad[GradIndex(0,0)] + Dgrad_grad[GradIndex(0,1)]/B - DB*grad_grad[GradIndex(0,1)]/B/B
				- 0.25*exp(-4.*u)*(Dx/B/B*grad_grad[GradIndex(2,2)] - 2.*DB*x/B/B/B*grad_grad[GradIndex(2,2)]
				                    + x/B/B*Dgrad_grad[GradIndex(2,2)] -4.*Du*x/B/B*grad_grad[GradIndex(2,2)]);

		
		if (idom == NDOM-1) {
			nu   = u  + log(B);
			Dnu  = Du + DB/B;
			vel  = sqrt(x)*B*exp(-2.*nu)*(Om-om);
			if (x > TINY) // x will only approach zero for the transition to spheroidal
			              // topologies for a vanishing Black Hole
				Dvel = exp(-2.*nu)*sqrt(x)*(0.5*Dx/x*B*(Om-om) + DB*(Om-om) - 2.*Dnu*B*(Om-om) + B*(DOm-Dom));
			else // Dx/sqrt(x)=0
				Dvel = 0.;
			vel2 = vel*vel;
			Dvel2= 2.*vel*Dvel;
			e2V            = exp(2.*nu)*(1.-vel2);
			De2V           = exp(2.*nu)*(2.*Dnu*(1.-vel2)-Dvel2);
			DV_func        = 0.5*De2V/e2V; // the name DV is already taken, hence DV_func
			h           = par.V0  - 0.5*log(e2V);
			Dh          = Dpar.V0 - 0.5*De2V/e2V;
			eos_quantities(  par, h, &p, &eps, &epsB);
			D_eos_quantities(par, h, Dh, &Dp, &Deps, &DepsB);

			JDX[Index(par,idom,0,i,j)] -= 4.*Pi*exp(2.*mu)*(
					2.*Dmu*((eps+p)*(1.+vel2)/(1.-vel2)-2.*p)
				+   (Deps+Dp)*(1.+vel2)/(1.-vel2) + (eps+p)*Dvel2/(1.-vel2) 
					                                            + (eps+p)*Dvel2*(1.+vel2)/(1.-vel2)/(1.-vel2)
					               - 2.*Dp
					);
			JDX[Index(par,idom,1,i,j)] -= 16.*Pi*exp(2.*mu)*(DB*p + B*Dp + 2.*Dmu*B*p);
			JDX[Index(par,idom,2,i,j)] -= -16.*Pi/sqrt(x)*exp(2.*mu+2.*u)*(
			     (-0.5*Dx/x + 2.*Dmu + 2.*Du)*B*(eps+p)*vel/(1.-vel2)
				 + DB*(eps+p)*vel/(1.-vel2) + B*(Deps+Dp)*vel/(1.-vel2) + B*(eps+p)*Dvel/(1.-vel2)
				 + B*(eps+p)*vel*Dvel2/(1.-vel2)/(1.-vel2)
			);
			JDX[Index(par,idom,3,i,j)] -= -4.*Pi*exp(2.*mu)*(2.*Dmu*(eps+p) + (Deps+Dp));
		}
	}
}
// -------------------------------------------------------------------------------
// -------------------------------------------------------------------------------
// -------------------------------------------------------------------------------
void NonLinRingBoundary(parameters par, derivs_1D w, derivs_1D G)
{// Calculating w.d0[Index_Ring(par,0,i)] from the known ring surface boundary condition,
 // see page 11, point 5.4
	int idom = NDOM-1, ns = par.ns[idom], Ns = ns-1, i, j=0;
	double V0 = par.V0, Om = par.Om;
	coordinate xx, yy;

	for(i=0;i<ns;i++){
		double a=Pih*i/Ns, sa=sin(a), s=sa*sa, t=0., B, W, v_e2nu, e2nu;

		coord_st_To_xy(par, idom, i, j, s, t, G, &xx, &yy);

		B  = w.d0[Index_Ring(par,1,i)];
		W  = sqrt(xx.d0)*B;
		v_e2nu	= W*(Om - w.d0[Index_Ring(par,2,i)]);  // v*e^(2*nu)=W*(Om-om)
		//   e^(2*V0) = e^(2*nu)*(1-v^2)  = e^(2*nu) - e^(-2*nu)*[v*e^(2*nu)]^2
		//=> e^(4*nu) - e^(2*V0)*e^(2*nu) - [v*e^(2*nu)]^2 = 0   
		e2nu	= 0.5*(exp(2.*V0) + sqrt(exp(4.*V0) + 4.*v_e2nu*v_e2nu));
		w.d0[Index_Ring(par,0,i)] = 0.5*log(e2nu) - log(B);
	}
}
// -------------------------------------------------------------------------------
void LinRingBoundary(parameters par, parameters Dpar,
         derivs_1D w, derivs_1D G, derivs_1D Dw, derivs_1D DG)
{// Linear version of 'NonLinRingBoundary'
	int idom = NDOM-1, ns = par.ns[idom], Ns = ns-1, i, j=0;
	double
		 V0 = par.V0,   Om = par.Om,
		DV0 = Dpar.V0, DOm = Dpar.Om;
	coordinate xx, yy, Dxx, Dyy; // why not simply x and y ????

	for(i=0;i<ns;i++){
		double a=Pih*i/Ns, sa=sin(a), s=sa*sa, t=0., B, DB, W, DW, v_e2nu, Dv_e2nu, e2nu, De2nu;

		Dcoord_st_To_xy(par, Dpar, idom, i, j, s, t, G, DG, &xx, &yy, &Dxx, &Dyy);

		B   = w.d0[Index_Ring(par,1,i)];    DB  = Dw.d0[Index_Ring(par,1,i)];
		W   = sqrt(xx.d0)*B;                DW  = sqrt(xx.d0)*DB + 0.5*Dxx.d0/sqrt(xx.d0)*B;
		v_e2nu	= W*(Om - w.d0[Index_Ring(par,2,i)]);  // v*e^(2*nu)=W*(Om-om)
		Dv_e2nu	= DW*(Om - w.d0[Index_Ring(par,2,i)]) + W*(DOm - Dw.d0[Index_Ring(par,2,i)]);
		//   e^(2*V0) = e^(2*nu)*(1-v^2)  = e^(2*nu) - e^(-2*nu)*[v*e^(2*nu)]^2
		//=> e^(4*nu) - e^(2*V0)*e^(2*nu) - [v*e^(2*nu)]^2 = 0   
		e2nu	= 0.5*(exp(2.*V0)        + sqrt(exp(4.*V0) + 4.*v_e2nu*v_e2nu));
		De2nu	= 0.5*(2.*DV0*exp(2.*V0) + 0.5*(4.*DV0*exp(4.*V0) + 8.*v_e2nu*Dv_e2nu)/sqrt(exp(4.*V0) + 4.*v_e2nu*v_e2nu));
		Dw.d0[Index_Ring(par,0,i)] = 0.5*De2nu/e2nu - DB/B;
	}
}
// -------------------------------------------------------------------------------
void NonLinInfinityEqns(parameters par, int idom, int i, int j, derivs_2D v, double *F)
{// Boundary conditions at infinity
	int indx;	
		
	indx = Index(par,idom,0,i,j); // Potential u, ipot=0
	F[indx] = v.d0[indx];
	
	indx = Index(par,idom,1,i,j); // Potential B, ipot=1
	F[indx] = v.d0[indx] - 1.;
	
	indx = Index(par,idom,2,i,j); // Potential om, ipot=2
	F[indx] = v.d0[indx];

	indx = Index(par,idom,3,i,j); // Potential mu, ipot=3
	F[indx] = v.d0[indx];
}	
// -------------------------------------------------------------------------------
void LinInfinityEqns(parameters par, int idom, int i, int j, derivs_2D Dv, double *JDX)
{// Linear version of 'NonLinInfinityEqns'
	int indx;	
		
	indx = Index(par,idom,0,i,j); // Potential u, ipot=0
	JDX[indx] = Dv.d0[indx];
	
	indx = Index(par,idom,1,i,j); // Potential B, ipot=1
	JDX[indx] = Dv.d0[indx];

	indx = Index(par,idom,2,i,j); // Potential om, ipot=2
	JDX[indx] = Dv.d0[indx];

	indx = Index(par,idom,3,i,j); // Potential mu, ipot=3
	JDX[indx] = Dv.d0[indx];

}
// -------------------------------------------------------------------------------
double NonLinU(parameters par, int idom, int ipot, int i, int j, derivs_2D v, derivs_1D w, derivs_1D G)
{// Like 'NonLinPotentials' but only calculates the value of the potential (ipot)
	int indx = Index(par,idom,ipot,i,j);
	double U;
	
	if (idom<NDOM-1)   U = v.d0[indx];
	else{
		int ns=par.ns[idom], Ns=ns-1, nt=par.nt[idom], Nt=nt-1,
			indx_R = Index_Ring(par,ipot,i);
		double 
			a=Pih*i/Ns, sa=sin(a), s=sa*sa,
			b=Pih*j/Nt, sb=sin(b), t=sb*sb;
		coordinate xx, yy;
		coord_st_To_xy(par, idom, i, j, s, t, G, &xx, &yy);
		U = w.d0[indx_R]  +  yy.d2*v.d0[indx];
	}
	
	return U;
}
// -------------------------------------------------------------------------------
double LinU(parameters par, parameters Dpar, int idom, int ipot, int i, int j, 
	derivs_2D v, derivs_1D w, derivs_1D G, derivs_2D Dv, derivs_1D Dw, derivs_1D DG)
{// Linear version of 'NonLinU'
	int indx = Index(par,idom,ipot,i,j);
	double DU;
	
	if (idom<NDOM-1)   DU = Dv.d0[indx];
	else{
		int ns=par.ns[idom], Ns=ns-1, nt=par.nt[idom], Nt=nt-1,
			indx_R = Index_Ring(par,ipot,i);
		double 
			a=Pih*i/Ns, sa=sin(a), s=sa*sa,
			b=Pih*j/Nt, sb=sin(b), t=sb*sb;
		coordinate xx, yy, Dxx, Dyy;
		Dcoord_st_To_xy(par, Dpar, idom, i, j, s, t, G, DG, &xx, &yy, &Dxx, &Dyy);
		DU = Dw.d0[indx_R]  + Dyy.d2*v.d0[indx] + yy.d2*Dv.d0[indx];
	}
	
	return DU;
}
// -------------------------------------------------------------------------------
void NonLinEquality(parameters par, int idom1, int idom2, int i1, int i2, int j1,  int j2, 
	derivs_2D v, derivs_1D w, derivs_1D G, derivs_2D V, double *F)
{// Difference of boundary function values at the same spatial grid point but with different indices
 // (idom1, i1, j1) != (idom2, i2, j2) 
 // to be used for continuity transition conditions
 // between regions 1 and 3 at the two corners s*(1-s)=0, t=1.
 // The differences are stored in F[Index(par,idom1,ipot,i1,j1)], ipot = 0 .. NPOT-1;
 // Here: idom1 = 3, j1 = par.nt[3]-1, idom2 = 1, j2 = par.nt[1]-1, i1=i2 \in [0;..;par.ns[1]-1]
	int indx1, indx2, ipot;
	double U1, U2;
	
	for(ipot=0; ipot<NPOT; ipot++) {
		indx1    = Index(par,idom1,ipot,i1,j1);
		indx2    = Index(par,idom2,ipot,i2,j2);
		U1       = NonLinU(par, idom1, ipot, i1, j1, v, w, G);
		U2       = NonLinU(par, idom2, ipot, i2, j2, v, w, G);
		F[indx1] = U1-U2;
	}
}
// -------------------------------------------------------------------------------
void LinEquality(parameters par, parameters Dpar, int idom1, int idom2, int i1, int i2, int j1,  int j2, 
	derivs_2D  v, derivs_1D  w, derivs_1D  G, derivs_2D  V, 
	derivs_2D Dv, derivs_1D Dw, derivs_1D DG, derivs_2D DV, double *JDX)
{// Linear version of 'LinU'
	int indx1, indx2, ipot;
	double DU1, DU2;
	
	for(ipot=0; ipot<NPOT; ipot++) {
		indx1      = Index(par,idom1,ipot,i1,j1);
		indx2      = Index(par,idom2,ipot,i2,j2);
		DU1        = LinU(par, Dpar, idom1, ipot, i1, j1, v, w, G, Dv, Dw, DG);
		DU2        = LinU(par, Dpar, idom2, ipot, i2, j2, v, w, G, Dv, Dw, DG);
		JDX[indx1] = DU1-DU2;
	}
}
// -------------------------------------------------------------------------------
void NonLinTransitEqns(parameters par, int idom, int idom1, int idom2, int is, int st, int i, int j, 
	derivs_2D v, derivs_1D w, derivs_1D G, derivs_2D V, double *F)
{// Transition equations:
 // 1) Calculating the (x,y)-coordinates for given (s,t) (and derivatives)
 // 2) dx,dy: Derivatives with respect to the boundary direction:
 //       if st == 1: Boundary is described by t=0 or t=1, s varies along boundary
 //       if st == 2: Boundary is described by s=0 or s=1, t varies along boundary
 // 3) Calculating U and the normal derivative U_n in (x,y)-coordinates for ipot=0..NPOT-1
 //       Exception: for {idom=NDOM-2=2 and s*(1-s)=0 and t=0} calculation of U_n*y_s
 // 4) Both function value and normal derivative is multiplied by (is)
 //    and added to the values
 //       F[Index(par,idom1,ipot,i1,j1)] and F[Index(par,idom2,ipot,i2,j2)] respectively,
 //    and the values (i1,j1) are determined from (i,j) with respect to the location on the boundary
	int ns=par.ns[idom], Ns=ns-1, nt=par.nt[idom], Nt=nt-1, ipot, indx1, indx2;
	double dx, dy, normal,
		a=Pih*i/Ns, sa=sin(a), s=sa*sa,
		b=Pih*j/Nt, sb=sin(b), t=sb*sb;
	coordinate xx, yy;

	coord_st_To_xy(par, idom, i, j, s, t, G, &xx, &yy);

	if (st==1){ 
		dx=xx.d1;
		dy=yy.d1;
	}
	else{
		dx=xx.d2;
		dy=yy.d2;
	}
	
	if (idom<NDOM-1){ 					// Calculation of (V, V_x, V_y), unless:
		v_To_V(par, idom, i, j, v, V); 	// {idom=NDOM-2=2 and s*(1-s)=0 and t=0}; then: (V, V_s, V_t)
		if (idom != NDOM-2 || i*(Ns-i) != 0 || j > 0) pot_st_To_xy(par, xx, yy, V);
	}
	else{
		double xs=xx.d1, ys=yy.d1, yt=yy.d2, yst=yy.d12, Vs;
		for(ipot=0; ipot<NPOT; ipot++) {
			int indx = Index(par,idom,ipot,i,j), indx_R = Index_Ring(par,ipot,i);
			Vs         = w.d1[indx_R]  + yst*v.d0[indx]  + yt*v.d1[indx];
			V.d0[ipot] = w.d0[indx_R]  +  yt*v.d0[indx];
			V.d2[ipot] = v.d2[indx];
			V.d1[ipot] = (Vs-ys*V.d2[ipot])/xs;
		}
	}
	
	for(ipot=0; ipot<NPOT; ipot++) {
		int i1=i, j1=j, i2=i, j2=j;

		if(st == 1 && j > 0) { j1=par.nt[idom1]-1; j2=par.nt[idom2]-1;}
		if(st == 2 && i > 0) { i1=par.ns[idom1]-1; i2=par.ns[idom2]-1;}

		indx1 = Index(par,idom1,ipot,i1,j1);
		indx2 = Index(par,idom2,ipot,i2,j2);

		F[indx1] += V.d0[ipot]*is;
		
		if (idom < NDOM-2 || j>0)
			normal = dy*V.d1[ipot] - dx*V.d2[ipot];
		else{ // For transition between regions 2 and 3, choose continuity of (y_s*normal)
			if (idom == NDOM-2 && i*(Ns-i) == 0 && j == 0 ) 
				normal = (dx*dx+dy*dy)*V.d2[ipot]/xx.d2-dx*V.d1[ipot];
			else
				normal = dy*(dy*V.d1[ipot] - dx*V.d2[ipot]);
		}		
		F[indx2]  += normal*is;
	}
}
// -------------------------------------------------------------------------------
void LinTransitEqns(parameters par, parameters Dpar, int idom, int idom1, int idom2, int is, int st, int i, int j, derivs_2D  v, derivs_1D  w, derivs_1D  G, derivs_2D  V,
          derivs_2D Dv, derivs_1D Dw, derivs_1D DG, derivs_2D DV, double* JDX)
{// Linear version of 'NonLinTransitEqns'

	int ns=par.ns[idom], Ns=ns-1, nt=par.nt[idom], Nt=nt-1, ipot, indx1, indx2;
	double dx, dy, Ddx, Ddy, Dnormal,
		a=Pih*i/Ns, sa=sin(a), s=sa*sa,
		b=Pih*j/Nt, sb=sin(b), t=sb*sb;
	coordinate xx, yy, Dxx, Dyy;

	Dcoord_st_To_xy(par, Dpar, idom, i, j, s, t, G, DG, &xx, &yy, &Dxx, &Dyy);
	if (st==1){ 
		dx=xx.d1;	Ddx=Dxx.d1;
		dy=yy.d1;	Ddy=Dyy.d1;
	}
	else{
		dx=xx.d2;	Ddx=Dxx.d2;
		dy=yy.d2;	Ddy=Dyy.d2;
	}
	
	if (idom<NDOM-1){ // Calculation of (V, V_x, V_y,; DV, DV_x, DV_y), unless:
		v_To_V(par, idom, i, j,  v,  V); 	// {idom=NDOM-2=2 & s*(1-s)=0 & t=0}; 
		v_To_V(par, idom, i, j, Dv, DV); 	// then: (V, V_s, V_t; DV, DV_s, DV_t)
		if (idom != NDOM-2 || i*(Ns-i) != 0 || j > 0) 
			D_pot_st_To_xy(par, xx, yy, V, Dxx, Dyy, DV);
	}
	else{
		double   xs=xx.d1,   ys=yy.d1,   yt=yy.d2,   yst=yy.d12,   Vs,
				Dxs=Dxx.d1, Dys=Dyy.d1, Dyt=Dyy.d2, Dyst=Dyy.d12, DVs;
		for(ipot=0; ipot<NPOT; ipot++) {
			int indx = Index(par,idom,ipot,i,j), indx_R = Index_Ring(par,ipot,i);
			V.d0[ipot]   = w.d0[indx_R]  +  yt*v.d0[indx];
			DV.d0[ipot]  = Dw.d0[indx_R] + Dyt*v.d0[indx]    + yt*Dv.d0[indx];
			Vs           = w.d1[indx_R]  + yst*v.d0[indx]    + yt*v.d1[indx];
			DVs          = Dw.d1[indx_R] + (Dyst*v.d0[indx]  + yst*Dv.d0[indx])
							+ (Dyt*v.d1[indx] + yt*Dv.d1[indx]);
			V.d2[ipot]   = v.d2[indx];
			DV.d2[ipot]  = Dv.d2[indx];
			V.d1[ipot]   = (Vs-ys*V.d2[ipot])/xs;
			DV.d1[ipot]  = (DVs-Dys*V.d2[ipot]-ys*DV.d2[ipot]-Dxs*V.d1[ipot])/xs;
		}
	}
	
	for(ipot=0; ipot<NPOT; ipot++) {
		int i1=i, j1=j, i2=i, j2=j;

		if(st == 1 && j > 0) { j1=par.nt[idom1]-1; j2=par.nt[idom2]-1;}
		if(st == 2 && i > 0) { i1=par.ns[idom1]-1; i2=par.ns[idom2]-1;}

		indx1 = Index(par,idom1,ipot,i1,j1);
		indx2 = Index(par,idom2,ipot,i2,j2);

		JDX[indx1] += DV.d0[ipot]*is;
		
		if (idom < NDOM-2 || j>0) 
			Dnormal = Ddy*V.d1[ipot]+dy*DV.d1[ipot] - Ddx*V.d2[ipot]-dx*DV.d2[ipot];
		else{ // For transition between regions 2 and 3, choose continuity of (y_s*normal)
			if (idom == NDOM-2 && i*(Ns-i) == 0 && j == 0 ){
				// normal = (dx*dx+dy*dy)*V.d2[ipot]/xx.d2-dx*V.d1[ipot];
				double 	aux0  = dx*dx+dy*dy, 
						Daux0 = 2.*(dx*Ddx+dy*Ddy),
						aux1  = aux0*V.d2[ipot]/xx.d2,
						Daux1 = (Daux0*V.d2[ipot]+aux0*DV.d2[ipot]-Dxx.d2*aux1)/xx.d2,
						Daux2 = Ddx*V.d1[ipot]+dx*DV.d1[ipot];
				Dnormal = Daux1-Daux2;
			}
			else{
				// normal = dy*(dy*V.d1[ipot] - dx*V.d2[ipot]);
				double 	aux0  = dy* V.d1[ipot] - dx*V.d2[ipot],
						Daux0 = Ddy*V.d1[ipot] + dy*DV.d1[ipot] - Ddx*V.d2[ipot] - dx*DV.d2[ipot];
				Dnormal = Ddy*aux0 + dy*Daux0;
			}
		}		
		JDX[indx2]  += Dnormal*is;
	}
}
// -------------------------------------------------------------------------------
void NonLinBoundEqns(parameters par, int idom, int i, int j, 
						derivs_2D v, derivs_1D w, derivs_1D G, derivs_2D V, double *F)
{// The Nonlinear Boundary Equations, see pages 13, 14
	int Ns=par.ns[idom]-1, Nt=par.nt[idom]-1;

	switch(idom){
		case 0:
			if (i==0) // s=0
				NonLinInfinityEqns(par, idom, i, j, v, F);
			else 
				if(i==Ns) // Transition between domains 0 and 1, values to be subtracted, boundary s=1
					NonLinTransitEqns(par, idom, 1, 0, -1, 2, i, j, v, w, G, V, F); 
				else
					NonLinFieldEqns(par, idom, i, j, v, w, G, V, F);
			break;
		case 1:
			if (i==Ns)   // Transition between domains 0 and 1, values to be added, boundary s=1
				NonLinTransitEqns(par, idom, 1, 0, +1, 2, i, j, v, w, G, V, F);
			else{
				if (j==Nt)// Transition between domains 1 and 2, values to be added, boundary t=1
					NonLinTransitEqns(par, idom, 2, 1, +1, 1, i, j, v, w, G, V, F); 
				else
					NonLinFieldEqns(par, idom, i, j, v, w, G, V, F);
			}
			break;
		case 2:
			if (j==Nt){
				if ((Ns-i)==0) // Continuity between region 1 and 2 enforced for s=1 and t=1
					NonLinEquality(par, 2, 1, i, i, par.nt[2]-1, par.nt[1]-1, v, w, G, V, F);
				else // Transition between domains 1 and 2, values to be substracted, boundary t=1
					NonLinTransitEqns(par, idom, 2, 1, -1, 1, i, j, v, w, G, V, F);
			}
			else{
				if (j==0)// Transition between domains 2 and 3, values to be added, boundary t=0
					NonLinTransitEqns(par, idom, 3, 2, +1, 1, i, j, v, w, G, V, F); 
				else // see page 13
					NonLinFieldEqns(par, idom, i, j, v, w, G, V, F);
			}
			break;
		case 3:
			if (j==0) // Transition between domains 2 and 3, values to be substracted, boundary t=0
				NonLinTransitEqns(par, idom, 3, 2, -1, 1, i, j, v, w, G, V, F);
			else // see page 13
				NonLinFieldEqns(par, idom, i, j, v, w, G, V, F);
			break;
	}
}
// -------------------------------------------------------------------------------
void LinBoundEqns(parameters par, parameters Dpar, int idom, int i, int j, 
					derivs_2D v, derivs_1D w, derivs_1D G, derivs_2D V, 
					derivs_2D Dv, derivs_1D Dw, derivs_1D DG, derivs_2D DV, double *JDX)
{// Linear Version of 'NonLinBoundEqns'
	int Ns=par.ns[idom]-1, Nt=par.nt[idom]-1;

	switch(idom){
		case 0:
			if (i==0)
				LinInfinityEqns(par, idom, i, j, Dv, JDX);
			else
				if(i==Ns)// Transition between domains 0 and 1, values to be subtracted, boundary s=1
					LinTransitEqns(par, Dpar, idom, 1, 0, -1, 2, i, j, v, w, G, V, Dv, Dw, DG, DV, JDX); 
				else
					LinFieldEqns(par, Dpar, idom, i, j, v, w, G, V, Dv, Dw, DG, DV, JDX);
			break;
		case 1:
			if (i==Ns)   // Transition between domains 0 and 1, values to be added, boundary s=1
				LinTransitEqns(par, Dpar, idom, 1, 0, +1, 2, i, j, v, w, G, V, Dv, Dw, DG, DV, JDX);
			else{
				if (j==Nt)// Transition between domains 1 and 2, values to be added, boundary s=0
					LinTransitEqns(par, Dpar, idom, 2, 1, +1, 1, i, j, v, w, G, V, Dv, Dw, DG, DV, JDX);
				else
					LinFieldEqns(par, Dpar, idom, i, j, v, w, G, V, Dv, Dw, DG, DV, JDX);
			}
			break;
		case 2:
			if (j==Nt){
				if (Ns-i==0) // Continuity between region 1 and 2 enforced for s=1 and t=1
					LinEquality(par, Dpar, 2, 1, i, i, par.nt[3]-1, par.nt[1]-1,
							v, w, G, V, Dv, Dw, DG, DV, JDX);
				else // Transition between domains 1 and 2, values to be substracted, boundary t=1
					LinTransitEqns(par, Dpar, idom, 2, 1, -1, 1, i, j, v, w, G, V, Dv, Dw, DG, DV, JDX);
			}
			else{
				if (j==0)// Transition between domains 2 and 3, values to be added, boundary t=0
					LinTransitEqns(par, Dpar, idom, 3, 2, +1, 1, i, j, v, w, G, V, Dv, Dw, DG, DV, JDX);
				else // see page 13
					LinFieldEqns(par, Dpar, idom, i, j, v, w, G, V, Dv, Dw, DG, DV, JDX);
			}
			break;
		case 3:
			if (j==0) // Transition between domains 2 and 3, values to be substracted, boundary t=0
				LinTransitEqns(par, Dpar, idom, 3, 2, -1, 1, i, j, v, w, G, V, Dv, Dw, DG, DV, JDX);
			else // see page 13
				LinFieldEqns(par, Dpar, idom, i, j, v, w, G, V, Dv, Dw, DG, DV, JDX);
			break;
	}
}
// -------------------------------------------------------------------------------
