#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <ctype.h>
#include <time.h>
#include "utilities.h"
#include "Ring.h"

//   -----------------------------------------------
int main() 
{
	int Newton_iter, i_Seq, ipar, idom, test;
	parameters parInit, par, parGoal, parOld;
	double *X0, *X, h;

	printf("\n");
	printf(" *************************************************\n");
	printf(" * Homogeneous Relativistic Rings *\n");
	printf(" *************************************************\n");

	ScanFiles(&parInit, &X0, &parGoal);
	parGoal.eps =1.;
	output(&parInit, X0);

	par = parGoal;
	par.grid_choice==0;
	parOld = parInit;
//	GetNewResolution(parInit, &par, &X);

	#define pI parInit
	#define pG parGoal
	for(i_Seq = 0; i_Seq <= par.n_Seq; i_Seq ++){
		printf(" *******************************************************\n");
		printf(" * Calculation of Sequence element No. \t %4d :        *\n", i_Seq);
		printf(" *******************************************************\n");
		if(par.n_Seq > 0) 	h = 1.*i_Seq/par.n_Seq;
		else					h = 0.;
		test = 0;
		par.nst[0]=0;
		for (idom=0; idom<NDOM; idom++){
			par.ns[idom] = pI.ns[idom] + h*(pG.ns[idom] - pI.ns[idom]);
			par.nt[idom] = pI.nt[idom] + h*(pG.nt[idom] - pI.nt[idom]);
			if (par.ns[idom] != pI.ns[idom] || par.nt[idom] != pI.nt[idom]) test=1;
			par.nst[idom+1] = par.nst[idom] + par.ns[idom]*par.nt[idom];
		}
		Get_Arrays_n_And_n_v(&par);
		if (test==1 || i_Seq==0){
			GetNewResolution(parOld, &par, &X0);
		}

		if (fabs(pG.eps-pI.eps)>TINY) par.eps      = pI.eps     + h*(pG.eps     - pI.eps);

		for(ipar=0; ipar<NPAR; ipar++){
			switch(par.choice[ipar]){
				case CHOICE_V0  		:	par.V0       = pI.V0      + h*(pG.V0      - pI.V0); 		break;
				case CHOICE_Om  		:	par.Om       = pI.Om      + h*(pG.Om      - pI.Om); 		break;
				case CHOICE_xi  		:	par.xi       = pI.xi      + h*(pG.xi      - pI.xi);		break;
				case CHOICE_xo  		:	par.xo       = pI.xo      + h*(pG.xo      - pI.xo);		break;
				case CHOICE_r_ratio		:	par.r_ratio  = pI.r_ratio + h*(pG.r_ratio - pI.r_ratio);	break;
				case CHOICE_beta_i  	:	par.beta_i   = pI.beta_i  + h*(pG.beta_i  - pI.beta_i);	break;
				case CHOICE_beta_o  	:	par.beta_o   = pI.beta_o  + h*(pG.beta_o  - pI.beta_o);	break;
				case CHOICE_M  			:	par.M          = pI.M       + h*(pG.M      - pI.M);		break;
				case CHOICE_J   		:	par.J          = pI.J       + h*(pG.J       - pI.J );		break;
			}
		}
//		par.x0 = pI.x0 + h*(pG.x0-pI.x0);
//		par.x1 = pI.x1 + h*(pG.x1-pI.x1);
		
//		par.x0 = 5.0*par.xo;
//		par.x1 = 0.5*par.xi;

		Newton_iter = newton(par, X0);

		if (Newton_iter==1){
			if (i_Seq == par.n_Seq) par.grid_choice = 1;
			output(&par, X0);
			PrintToFile(par, X0, i_Seq);
			parOld=par;
		}
		else{
			printf("The Newton-Raphson routine failed to converge\n");
			break;
		}
	}
	#undef pI
	#undef pG

//	free_dvector(X, 0, par.ntotal-1);

	Free_Arrays_n_And_n_v(&par);

	return 1;
}
//   -----------------------------------------------
void output(parameters *par, double *X)
{   // setting and screen output of physical parameters
	int ns = (*par).ns[NDOM-1];
	double dummy, p_max;
	derivs_2D v; 
	derivs_1D G, w; 

	allocate_derivs_1D(&G, ns);
	allocate_derivs_1D(&w, ns*NPOT);
	allocate_derivs_2D(&v, (*par).n_v_total);

	Get_vwGpar_From_X(*par, X, v, w, G, par);

	NonLinRingBoundary(*par, w, G);
	Derivatives_s(*par, w);
	GetPhysicalParameters_NonLin(par, v, w, G);
	Volume_integrals(*par, v, w, G, &((*par).M0), &((*par).M_vol), &((*par).J_vol));
	eos_quantities(*par, (*par).h_max, &p_max, &dummy, &dummy);

	printf(" \n");
	printf(" Parameters:\n");
	printf(" -----------\n\n");
	printf(" M       = %20.15e \t J       = %20.15e \n", (*par).M, (*par).J);
//	printf(" M_infty = %20.15e \t J_infty = %20.15e \n", (*par).M_infty, (*par).J_infty);
//	printf(" M_in    = %20.15e \t J_in    = %20.15e \n", (*par).M_vol, (*par).J_vol);
	printf(" M0      = %20.15e \t Ebind   = %20.15e\n", (*par).M0,  (*par).M0-(*par).M);
	printf(" Rcirc_i = %20.15e \t Rcirc_o = %20.15e\n\n", (*par).Rcirc_i,  (*par).Rcirc_o);

	printf(" z       = %20.15e \t V0      = %20.15e \t  \n", exp(-(*par).V0)-1.,(*par).V0);
	printf(" rho_i   = %20.15e \t rho_o   = %20.15e \n", sqrt((*par).xi),  sqrt((*par).xo));
	printf(" r_ratio = %20.15e \t Omega   = %20.15e \n", (*par).r_ratio, (*par).Om);
	printf(" beta_i  = %20.15e \t beta_o  = %20.15e \n", (*par).beta_i, (*par).beta_o);
	printf(" p_max   = %20.15e \t h_max  = %20.15e \n\n", p_max , exp((*par).h_max));

	printf(" x0      = %20.15e \t x1      = %20.15e \n\n", X0FACT*(*par).xo, X1FACT*(*par).xi);
  printf("Tests: |1-M_infty/M|   = %20.15e \t |1-J_infty/J|   = %20.15e\n\n",
    fabs(1.-(*par).M_infty/(*par).M),fabs(1.-(*par).J_infty/(*par).J));
	printf(" number of coefficients in each direction and each domain= %d  \n", (*par).ns[0]);


	if ((*par).grid_choice==1){
		surf_file(*par, G);
		ergosphere_file(*par, v, w, G);
		print_grid(*par, v, w, G);
	}

	free_derivs_1D(&G, ns);
	free_derivs_1D(&w, ns*NPOT);
	free_derivs_2D(&v, (*par).n_v_total);
}
	
