#define NDOM 				4
#define NPOT 				4
#define NPAR 				2

#define CHOICE_M 		 1
#define CHOICE_J 		 2
#define CHOICE_xi 		 3
#define CHOICE_xo 		 4
#define CHOICE_r_ratio 		 5
#define CHOICE_beta_i	 	 6
#define CHOICE_beta_o		7
#define CHOICE_V0			8
#define CHOICE_Om			9

#define KNEUTRON 0.0111803398875   // =1/(8*5^1.5)
#define NMAX 				100

#define X0FACT 4.0
#define X1FACT 0.5

typedef struct PARAMETERS{ 
	double	//x0, x1,
		V0, Om, eps, B, B60_factor, alpha, K_poly, N_poly, xi, xo, Rcirc_i, Rcirc_o, r_ratio, beta_i, beta_o,
		M_infty, J_infty, M, J, M_vol, J_vol, M0, VI2, VI3,
		Newton_tol, p_max, h_max,
		bicgstab_decr, bicgstab_tol,
		PreCond_Improve, xmin, ymin, xmax, ymax;
	int n1, n2, ntotal, n_v_total, n_2D, n_v_2D,
		ns[NDOM], nt[NDOM], nst[NDOM+1], *n_of_n_v , *n_v_of_n, choice[NPAR], n_Seq,
		Newton_itmin, Newton_itmax, Newton_verb,
		bicgstab_itmax, bicgstab_verb,
		PreCond_itmax, PreCond_FD_Order, eos, nx, ny, grid_choice; // K and n are for polytropic eos
	char SeqName[200];
} parameters;

typedef struct COORDINATE{ 
	double d0, d1, d2, d11, d12, d22, d112;
} coordinate;

typedef struct ST{ 
	double s, t;
	int idom;
} st;

typedef struct DERIVS_1D{
	double *d0, *d1, *d11;
} derivs_1D;


typedef struct DERIVS_2D{ 
	double *d0, *d1, *d2, *d11, *d12, *d22;
} derivs_2D;

typedef struct JFD_COMPONENTS{ 
	double 
		**J_UL, **K_UL, **Kl_UL, **Z_UL, **H_UL, 
		**J_UR, **K_UR, 
		**J_LL, 
		**J_LR, **K_LR;
	int    
		*ncols_J_UL, **cols_J_UL, *iK_UL, *iZ_UL, *iH_UL, 
		*iK_LR, 
		m1, m2, col_Z_max;
} JFD_Components;

typedef struct P_EPS{
	double	p, eps;
} p_eps;

// Subroutines in "Ring.c"
int main();
void output(parameters *par, double *X);

// Subroutines in "ParameterEqns.c"
void M_and_J(parameters par, derivs_2D v, derivs_1D w, derivs_1D G, int ij, double *M, double *J);
void Volume_integrals(parameters par, derivs_2D v, derivs_1D w, derivs_1D G, double *M0, double *M, double *J);
double h_max(parameters par, derivs_2D v, derivs_1D w, derivs_1D G);
double Rcirc(parameters par, derivs_2D v, derivs_1D w, derivs_1D G, int i);
void GetPhysicalParameters_NonLin(parameters *parX, derivs_2D v, derivs_1D w, derivs_1D G);
void GetPhysicalParameters_Lin(parameters *parX, parameters *DparX, derivs_2D v, derivs_1D w, derivs_1D G, derivs_2D V, 
					derivs_2D Dv, derivs_1D Dw, derivs_1D DG, derivs_2D DV);
void NonLinParamEqns(parameters par, parameters parX, derivs_2D v, derivs_1D w, derivs_1D G, derivs_2D V, double *FF);
void LinParamEqns(parameters par, parameters parX, parameters DparX, 
					derivs_2D v, derivs_1D w, derivs_1D G, derivs_2D V, 
					derivs_2D Dv, derivs_1D Dw, derivs_1D DG, derivs_2D DV, double *JJDX);

// Subroutines in "pot_st_To_xy.c"
void pot_st_To_xy(parameters par, coordinate x, coordinate y, derivs_2D V);
void D_pot_st_To_xy(parameters par, coordinate x,  coordinate y,   derivs_2D V,
									coordinate Dx, coordinate Dy,  derivs_2D DV);

// Subroutines in "CoordTransf.c"
void coord_st_To_xy(parameters par, int idom, int i, int j, double s, double t, derivs_1D G, 
						coordinate *x, coordinate *y);
void Dcoord_st_To_xy(parameters par, parameters Dpar, int idom, int i, int j, double s, double t, 
	derivs_1D G, derivs_1D DG, coordinate *x, coordinate *y, coordinate *Dx, coordinate *Dy);
double Func_xm(parameters par);
double Func_Dxm(parameters par, parameters Dpar);
void xys_To_xy(double xm, coordinate xs, coordinate ys, coordinate *x, coordinate *y);
void D_xys_To_xy(double xm, double Dxm, coordinate xs, coordinate ys, coordinate Dxs, coordinate Dys, 
	coordinate *x, coordinate *y, coordinate *Dx, coordinate *Dy);
void xy_To_xys(double xm, coordinate x, coordinate y, coordinate *xs, coordinate *ys);
void D_xy_To_xys(double xm, double Dxm, coordinate x, coordinate y, coordinate Dx, coordinate Dy, 
	coordinate *xs, coordinate *ys, coordinate *Dxs, coordinate *Dys);
void coord_0(parameters par, int i, int j, double s, double t, derivs_1D G, 
						coordinate *x, coordinate *y);
void Dcoord_0(parameters par, parameters Dpar, int i, int j, double s, double t, 
	derivs_1D G, derivs_1D DG, coordinate *x, coordinate *y, coordinate *Dx, coordinate *Dy);
void coord_1(parameters par, int i, int j, double s, double t, derivs_1D G, 
						coordinate *x, coordinate *y);
void Dcoord_1(parameters par, parameters Dpar, int i, int j, double s, double t, 
	derivs_1D G, derivs_1D DG, coordinate *x, coordinate *y, coordinate *Dx, coordinate *Dy);
void coord_2(parameters par, int i, int j, double s, double t, derivs_1D G, 
						coordinate *x, coordinate *y);
void Dcoord_2(parameters par, parameters Dpar, int i, int j, double s, double t, 
	derivs_1D G, derivs_1D DG, coordinate *x, coordinate *y, coordinate *Dx, coordinate *Dy);
void coord_2_all_st(parameters par, int i, int j, double s, double t, derivs_1D G, coordinate *x, coordinate *y);
void coord_3(parameters par, int i, int j, double s, double t, derivs_1D G, 
						coordinate *x, coordinate *y);
void Dcoord_3(parameters par, parameters Dpar, int i, int j, double s, double t, 
	derivs_1D G, derivs_1D DG, coordinate *x, coordinate *y, coordinate *Dx, coordinate *Dy);
void coord_3_all_st(parameters par, double s, double t, derivs_1D Gcheb, coordinate *x, coordinate *y);
void coord_3_all_st_old(parameters par, double s, double t, derivs_1D G, coordinate *x, coordinate *y);
st coord_xy_To_st_old(parameters par, derivs_1D G, double x, double y);
st coord_xy_To_st(parameters par, derivs_1D G, derivs_1D Gcheb, double x, double y);
coordinate coordSum(coordinate a, coordinate b);
coordinate coordDiff(coordinate a, coordinate b);
coordinate coordInverse(coordinate a);
coordinate coordProduct(coordinate a, coordinate b);
coordinate coordQuotient(coordinate a, coordinate b);
coordinate coordRoot(coordinate a);
coordinate coordLog(coordinate a);
coordinate coordExp(coordinate a);
coordinate const_times_coord(double b, coordinate a);
coordinate const_plus_coord(double b, coordinate a);
coordinate DcoordQuotient(coordinate a, coordinate b, coordinate Da, coordinate Db);
coordinate DcoordProduct(coordinate a, coordinate b, coordinate Da, coordinate Db);
coordinate DcoordRoot(coordinate a, coordinate Da);
coordinate DcoordLog(coordinate a, coordinate Da);
coordinate DcoordExp(coordinate a, coordinate Da);
coordinate D_const_times_coord(double b, double Db, coordinate a, coordinate Da);

// Subroutines in "FuncAndJacobian.c"
int Index(parameters par, int idom, int ipot, int i, int j);
int Index_i(parameters par, int i);
int Index_Ring(parameters par, int ipot, int i);
void Get_Indices_From_n_v(parameters par, int n_v, int *idom, int *ipot, int *i, int *j, int *ipar);
void v_To_V(parameters par, int idom, int i, int j, derivs_2D v, derivs_2D V);
void Derivatives_st(parameters par, derivs_2D v);
void Derivatives_s(parameters par, derivs_1D w);
void Get_vwGpar_From_X(parameters par, double *X, derivs_2D v, derivs_1D w, derivs_1D G, parameters *parX);
void F_of_X(parameters par, double *X, derivs_2D v, derivs_1D w, derivs_1D G, parameters *parX, double *F);
void J_times_DX(parameters par, double *X, double *DX, derivs_2D v, derivs_1D w, derivs_1D G, parameters parX, double *JDX);

// Subroutines in "FieldAndBoundEqns.c"
int GradIndex(int i, int j);
void NonLinPotentials(parameters par, int idom, int i, int j, double s, double t, 
	coordinate xx, coordinate yy, derivs_2D v, derivs_1D w, derivs_1D G, derivs_2D V, double *Laplace,
	double *grad_grad);
void LinPotentials(parameters par, int idom, int i, int j, double s, double t, 
	coordinate  xx,  coordinate yy, derivs_2D v,  derivs_1D w,  derivs_1D G,  derivs_2D V,  double *Laplace, double *grad_grad,
	coordinate Dxx, coordinate Dyy, derivs_2D Dv, derivs_1D Dw, derivs_1D DG, derivs_2D DV, double *DLaplace,double *Dgrad_grad);
void NonLinFieldEqns(parameters par, int idom, int i, int j, 
	derivs_2D v, derivs_1D w, derivs_1D G, derivs_2D V, double *F);
void LinFieldEqns(parameters par, parameters Dpar, int idom, int i, int j, 
	derivs_2D  v, derivs_1D  w, derivs_1D  G, derivs_2D  V,
	derivs_2D Dv, derivs_1D Dw, derivs_1D DG, derivs_2D DV, double *JDX);
void NonLinRingBoundary(parameters par, derivs_1D w, derivs_1D G);
void LinRingBoundary(parameters par, parameters Dpar, derivs_1D w, derivs_1D G, derivs_1D Dw, derivs_1D DG);
void NonLinInfinityEqns(parameters par, int idom, int i, int j, derivs_2D v, double *F);
void LinInfinityEqns(parameters par, int idom, int i, int j, derivs_2D Dv, double *JDX);
double NonLinU(parameters par, int idom, int ipot, int i, int j, derivs_2D v, derivs_1D w, derivs_1D G);
double LinU(parameters par, parameters Dpar, int idom, int ipot, int i, int j, 
	derivs_2D v, derivs_1D w, derivs_1D G, derivs_2D Dv, derivs_1D Dw, derivs_1D DG);
void NonLinEquality(parameters par, int idom1, int idom2, int i1, int i2, int j1,  int j2, 
	derivs_2D v, derivs_1D w, derivs_1D G, derivs_2D V, double *F);
void LinEquality(parameters par, parameters Dpar, int idom1, int idom2, int i1, int i2, int j1,  int j2, 
	derivs_2D  v, derivs_1D  w, derivs_1D  G, derivs_2D  V, 
	derivs_2D Dv, derivs_1D Dw, derivs_1D DG, derivs_2D DV, double *JDX);
void NonLinTransitEqns(parameters par, int idom, int idom1, int idom2,  int is, int st, int i, int j, 
	derivs_2D v, derivs_1D w, derivs_1D G, derivs_2D V, double *F);
void LinTransitEqns(parameters par, parameters Dpar, int idom, int idom1, int idom2, int is, int st, int i, int j, 
						derivs_2D  v, derivs_1D  w, derivs_1D  G, derivs_2D  V, 
						derivs_2D Dv, derivs_1D Dw, derivs_1D DG, derivs_2D DV, double* JDX);
void NonLinBoundEqns(parameters par, int idom, int i, int j, 
						derivs_2D v, derivs_1D w, derivs_1D G, derivs_2D V, double *F);
void LinBoundEqns(parameters par, parameters Dpar, int idom, int i, int j, 
					derivs_2D v, derivs_1D w, derivs_1D G, derivs_2D V, 
					derivs_2D Dv, derivs_1D Dw, derivs_1D DG, derivs_2D DV, double *JDX);

// Subroutines in "JacobianFD.c"
void GetFD_Dv_a(parameters par, derivs_2D Dv, int idom, int ipot, int i, int j);
void GetFD_Dv_b(parameters par, derivs_2D Dv, int idom, int ipot, int i, int j);
void GetFD_Dv_ab_FD2(parameters par, derivs_2D Dv, int idom, int ipot, int i, int j);
void GetFD_Dv_ab_FD4(parameters par, derivs_2D Dv, int idom, int ipot, int i, int j);
void GetFD_Dv_st(parameters par, derivs_2D Dv, int idom, int ipot, int i, int j);
void GetFD_DG_s(parameters par, derivs_1D DG, int i);
void GetFD_Dw_s(parameters par, derivs_1D Dw, int ipot, int i);
void GetFD_DvwGpar(parameters par, int n_v, double *DX, derivs_2D v, derivs_1D w, derivs_1D G, 
	derivs_2D Dv, derivs_1D Dw, derivs_1D DG, parameters *DparX);
void fill0_Dv_of_indx(int indx, derivs_2D Dv);
void fill0_DvwGpar(parameters par, int n_v, derivs_2D Dv, derivs_1D Dw, derivs_1D DG, parameters *DparX);
void Get_boundary_i0(parameters par, int jdom, int *adom, int *ai0, int *ai1, int *aj0, int *aj1, int *mdom);
void Get_boundary_i1(parameters par, int jdom, int *adom, int *ai0, int *ai1, int *aj0, int *aj1, int *mdom);
void Get_boundary_j0(parameters par, int jdom, int *adom, int *ai0, int *ai1, int *aj0, int *aj1, int *mdom);
void Get_boundary_j1(parameters par, int jdom, int *adom, int *ai0, int *ai1, int *aj0, int *aj1, int *mdom);
void Get_gridpoints(parameters par, int idom, int ipot, int i, int j, int *mdom, int *adom, int *ai0, int *ai1, int *aj0, int *aj1);
void Get_JFD_Matrix(parameters par, double *X, derivs_2D v, derivs_1D w, derivs_1D G, parameters parX, JFD_Components *JFD);
void Get_JFD_Components(parameters par, JFD_Components *JFD);
						
// Subroutines in "newton.c"
void resid(parameters par, double *DX, double *rhs, JFD_Components JFD, double *res);
void Solve_JFD_UL(parameters par, double *F, double *X, JFD_Components JFD);
void Solve_JFD(parameters par, double *F, double *DX, JFD_Components JFD);
int PreCond(parameters par, double *F, double *DX, double *norm, JFD_Components JFD);
int bicgstab_Iterations(parameters par, double *X, derivs_2D v, derivs_1D w, derivs_1D G, parameters parX, 
		double *F, double *DX, double *normres, JFD_Components JFD);
int bicgstab(parameters par, double *X, derivs_2D v, derivs_1D w, derivs_1D G, parameters parX, double *F, double *DX, double *normres);
int newton(parameters par, double *X);


// Subroutines in "ScanAndPrint.c"
void Get_Arrays_n_And_n_v_domain_0(parameters *par, int *n);
void Get_Arrays_n_And_n_v_domain_1(parameters *par, int *n);
void Get_Arrays_n_And_n_v_domain_2(parameters *par, int *n);
void Get_Arrays_n_And_n_v_domain_3(parameters *par, int *n);
void Get_Arrays_n_And_n_v_domain_4(parameters *par, int *n);
void Get_Arrays_n_And_n_v(parameters *par);
void Free_Arrays_n_And_n_v(parameters *par);
void ScanFiles(parameters *parInit, double **X, parameters *parGoal);
void PrintToFile(parameters par, double *X, int i_Seq);

// Subroutines in "GetNewResolution.c"
void v1_To_v2(parameters par1, derivs_2D v1, parameters par2, derivs_2D v2);
void wG1_To_wG2(parameters par1, derivs_1D w1, derivs_1D G1, parameters par2, derivs_1D w2, derivs_1D G2);
void GetNewResolution(parameters par1, parameters *par2, double **X1);

// Subroutines in "allocate_and_free_derivs.c"
void allocate_derivs_1D(derivs_1D *v, int n);
void allocate_derivs_2D(derivs_2D *v, int n);
void fill0_derivs_1D(derivs_1D v, int n);
void fill0_derivs_2D(derivs_2D v, int n);
void free_derivs_1D(derivs_1D *v, int n);
void free_derivs_2D(derivs_2D *v, int n);

// Subroutines in "eqstate.c"
void scan_eos_parameters( parameters *par, FILE *stream);
void print_eos_parameters(parameters  par, FILE *stream);
void eos_quantities(parameters par, double h, double *p, double *eps, double *epsB);
void D_eos_quantities(parameters par, double h, double Dh, double *Dp, double *Deps, double *DepsB);
//p_eps p_e(parameters par,double eV0mV);
//p_eps Dp_e(parameters par, parameters Dpar, double eV0mV, double DV0mV);

// Subroutines in "surf_and_ergosphere.c"
void surf_file(parameters par, derivs_1D G);
void ergosphere_file(parameters par, derivs_2D v, derivs_1D w, derivs_1D G);
