#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <ctype.h>
#include <time.h>
#include "utilities.h"
#include "Ring.h"

// Scanning and printing data files.
// Also: The Index arrays n[n_v] and n_v[n} are determined

// -------------------------------------------------------------------------------
void Get_Arrays_n_And_n_v_domain_0(parameters *par, int *n)
{	
	int idom=0, ns=(*par).ns[idom], Ns=ns-1, nt=(*par).nt[idom], Nt=nt-1, i, j, ipot, n_v; 

	for(i=0;i<=Ns;i++){
		for(j=0;j<=Nt;j++) {
			for(ipot=0; ipot<NPOT; ipot++){
				n_v = Index(*par, idom, ipot, i, j);
				(*par).n_v_of_n[*n]  = n_v;
				(*par).n_of_n_v[n_v] = *n;
				*n += 1;
			}
		}
	}
}
// -------------------------------------------------------------------------------
void Get_Arrays_n_And_n_v_domain_1(parameters *par, int *n)
{	
	int idom=1, ns=(*par).ns[idom], Ns=ns-1, nt=(*par).nt[idom], Nt=nt-1, i, j, ipot, n_v; 

	i=Ns;
	for(j=0;j<=Nt;j++) { // boundary between domains 0 and 1
		for(ipot=0; ipot<NPOT; ipot++){
			n_v = Index(*par, idom, ipot, i, j);
			(*par).n_of_n_v[n_v] = -1;
		}
	}
	for(i=Ns-1;i>=0;i--){
		for(j=0;j<=Nt;j++) {
			for(ipot=0; ipot<NPOT; ipot++){
				n_v = Index(*par, idom, ipot, i, j);
				(*par).n_v_of_n[*n]  = n_v;
				(*par).n_of_n_v[n_v] = *n;
				*n += 1;
			}
		}
	}
}
// -------------------------------------------------------------------------------
void Get_Arrays_n_And_n_v_domain_2(parameters *par, int *n)
{	
	int idom=2, ns=(*par).ns[idom], Ns=ns-1, nt=(*par).nt[idom], Nt=nt-1, i, j, ipot, n_v; 

	for(j=Nt;j>=0;j--) {
		for(i=0;i<=Ns;i++){
			for(ipot=0; ipot<NPOT; ipot++){
				n_v = Index(*par, idom, ipot, i, j);
				(*par).n_v_of_n[*n]  = n_v;
				(*par).n_of_n_v[n_v] = *n;
				*n += 1;
			}
		}
	}
}
// -------------------------------------------------------------------------------
void Get_Arrays_n_And_n_v_domain_3(parameters *par, int *n)
{	
	int idom=3, ns=(*par).ns[idom], Ns=ns-1, nt=(*par).nt[idom], Nt=nt-1, i, j, ipot, n_v; 

	for(j=1;j<=Nt;j++) {
		for(i=0;i<=Ns;i++){
			for(ipot=0; ipot<NPOT; ipot++){
				n_v = Index(*par, idom, ipot, i, j);
				(*par).n_v_of_n[*n]  = n_v;
				(*par).n_of_n_v[n_v] = *n;
				*n += 1;
			}
		}
	}
	(*par).n2 = *n;

	j=0;
	for(i=0;i<=Ns;i++){
		for(ipot=0; ipot<NPOT; ipot++){
			n_v = Index(*par, idom, ipot, i, j);
			(*par).n_v_of_n[*n]  = n_v;
			(*par).n_of_n_v[n_v] = *n;
			*n += 1;
		}
	}
	(*par).n_2D = *n;
}
// -------------------------------------------------------------------------------
void Get_Arrays_n_And_n_v(parameters *par)
{
	int n=0, n_v, ipar;

	(*par).n_v_2D    = NPOT*(*par).nst[NDOM];
	(*par).n_v_total = (*par).n_v_2D + NPAR;
	
	(*par).n_of_n_v = ivector(0, (*par).n_v_total-1);
	(*par).n_v_of_n = ivector(0, (*par).n_v_total-1);

	Get_Arrays_n_And_n_v_domain_0(par, &n);// "start" integer n
	Get_Arrays_n_And_n_v_domain_1(par, &n);// provided and
	Get_Arrays_n_And_n_v_domain_2(par, &n);// "end" integer
	Get_Arrays_n_And_n_v_domain_3(par, &n);// par.n_2D retuned
//	Get_Arrays_n_And_n_v_domain_4(par, &n);// 
	(*par).ntotal = (*par).n_2D + NPAR;

	for(ipar=0; ipar<NPAR; ipar++){
		n   = (*par).n_2D    + ipar;
		n_v = (*par).n_v_2D  + ipar;
		(*par).n_v_of_n[n]   = n_v;
		(*par).n_of_n_v[n_v] = n;
	}
}
// -------------------------------------------------------------------------------
void Free_Arrays_n_And_n_v(parameters *par)
{
	free_ivector((*par).n_of_n_v, 0, (*par).n_v_total-1);
	free_ivector((*par).n_v_of_n, 0, (*par).n_v_total-1);
}
//   -----------------------------------------------
void ScanInitialData(parameters *parInit, double **X)
{  /* A file called "InitialData" is read.
     Information is stored in parInit (essentially bookkeeping)
	  Physical information is stored in X 
	*/ 
	FILE *stream_init;
	char string[200];
	int n, n_v, idom, ipot, ipar, ns, i, nt, j;
	double d_aux;	// d_aux: a double auxiliary variable

	stream_init   = fopen("InitialData", "r");
	
	(*parInit).nst[0] = 0;
	for(idom=0; idom<NDOM; idom++){
		fscanf(stream_init, " %d %d %s ",  &ns, &nt, string);
		(*parInit).ns[idom] = ns;
		(*parInit).nt[idom] = nt;
		(*parInit).nst[idom+1] = (*parInit).nst[idom] + ns*nt;
	}
	Get_Arrays_n_And_n_v(parInit);

	fscanf(stream_init, " %lf %s ",  &d_aux, string);//		(*parInit).x0  = d_aux;
	fscanf(stream_init, " %lf %s ",  &d_aux, string);//		(*parInit).x1  = d_aux;
	scan_eos_parameters(parInit, stream_init);

	*X  = dvector(0, (*parInit).ntotal-1);
	for(idom=0; idom<NDOM; idom++){
		ns=(*parInit).ns[idom];
		nt=(*parInit).nt[idom];
		for(ipot=0; ipot<NPOT; ipot++){
			for(j=0;j<nt;j++) {
				for(i=0;i<ns;i++){
					n_v = Index(*parInit, idom,ipot,i,j);
					n   = (*parInit).n_of_n_v[n_v];
					if(n>=0){
						fscanf(stream_init, " %lf %s ", &d_aux, string);
						(*X)[n] = d_aux;
					}
				}
			}
		}
	}
	for(ipar=0; ipar<NPAR; ipar++){
		fscanf(stream_init, " %lf %s ", &d_aux, string);
		(*X)[(*parInit).n_2D + ipar] = d_aux;
//printf("ipar=%d \t X=%e\n",ipar,d_aux); getchar();
	}
	fclose(stream_init);
}
//   -----------------------------------------------
void ScanConfig(parameters *parInit, parameters *parGoal)
{
	FILE *stream_config;
	char string[200];
	int i_aux, idom, ipar, n, nx, ny;	// i_aux: an int auxiliary variable 
	double d_aux, rhomin, zetamin, rhomax, zetamax;							// d_aux: a double auxiliary variable

	stream_config = fopen("Config", "r");

	fscanf(stream_config, " %s  %s",  (*parInit).SeqName, string);

	(*parInit).Newton_itmin     = 0;
	fscanf(stream_config, " %d  %s ",  &i_aux, string);		(*parInit).Newton_itmax     = i_aux;
	fscanf(stream_config, " %lf %s ",  &d_aux, string);		(*parInit).Newton_tol       = d_aux;
	(*parInit).Newton_verb      = 1;
	
	(*parInit).bicgstab_itmax   = 40;
	(*parInit).bicgstab_decr    = 1.e-5;
	(*parInit).bicgstab_verb    = 0;
	
	(*parInit).PreCond_Improve  = 1.e-12;
	(*parInit).PreCond_itmax    = 5;
	(*parInit).PreCond_FD_Order = 4;

	*parGoal = *parInit;
	(*parGoal).nst[0] = 0;
	fscanf(stream_config, " %d %s ",  &n, string);
	for(idom=0; idom<NDOM; idom++){
		(*parGoal).ns[idom] = n;
		(*parGoal).nt[idom] = n;
		(*parGoal).nst[idom+1] = (*parGoal).nst[idom] + n*n;
	}
	Get_Arrays_n_And_n_v(parGoal);

	fscanf(stream_config, " %s ",  string);
	for(ipar=0; ipar<NPAR; ipar++){
		fscanf(stream_config, " %d  %s ",  &i_aux, string); 
		(*parInit).choice[ipar] = (*parGoal).choice[ipar] = i_aux;
	}

	for(ipar=0; ipar<NPAR; ipar++){
		fscanf(stream_config, " %lf %s ",  &d_aux, string);
		switch((*parGoal).choice[ipar]){
			case CHOICE_V0  		:	(*parGoal).V0       = d_aux;	break;
			case CHOICE_Om 			:	(*parGoal).Om       = d_aux;	break;
			case CHOICE_xi  		:	(*parGoal).xi       = d_aux;	break;
			case CHOICE_xo  		:	(*parGoal).xo       = d_aux;	break;
			case CHOICE_r_ratio		:	(*parGoal).r_ratio  = d_aux;	break;
			case CHOICE_beta_i  	:	(*parGoal).beta_i   = d_aux;	break;
			case CHOICE_beta_o  	:	(*parGoal).beta_o   = d_aux;	break;
			case CHOICE_M	  		:	(*parGoal).M       = d_aux;	break;
			case CHOICE_J  		:	(*parGoal).J       = d_aux;	break;
		}
	}

	fscanf(stream_config, " %d  %s ",  &i_aux, string); (*parGoal).n_Seq = i_aux;

	fscanf(stream_config, " %d %d %s ",  &nx, &ny, string);
	(*parInit).nx   = (*parGoal).nx   = nx;
	(*parInit).ny   = (*parGoal).ny   = ny;
	fscanf(stream_config, " %lf %lf %s ",  &rhomin, &zetamin, string);  
	(*parInit).xmin = (*parGoal).xmin = rhomin*rhomin;
	(*parInit).ymin = (*parGoal).ymin = zetamin*zetamin;
	fscanf(stream_config, " %lf %lf %s ",  &rhomax, &zetamax, string);  
	(*parInit).xmax = (*parGoal).xmax = rhomax*rhomax;
	(*parInit).ymax = (*parGoal).ymax = zetamax*zetamax;


	scan_eos_parameters(   parGoal, stream_config);

//	fscanf(stream_config, " %lf %s ",  &d_aux, string);	(*parGoal).x0    = d_aux;
//	fscanf(stream_config, " %lf %s ",  &d_aux, string);	(*parGoal).x1    = d_aux;

	fclose(stream_config);
}
//   -----------------------------------------------
void ScanFiles(parameters *parInit, double **X, parameters *parGoal)
{
	ScanInitialData(parInit, X);
	ScanConfig(parInit, parGoal);
}
//   -----------------------------------------------
void PrintToFile(parameters par, double *X, int i_Seq)
{
	char name[50];
	FILE *stream;
	int n_2D=par.n_2D, idom, ipot, i, j, ns, Ns, nt, Nt, n, n_v;
	
	if (i_Seq <  10)              sprintf(name, "%s.00%1d", par.SeqName, i_Seq);
	if (i_Seq >= 10 && i_Seq<100) sprintf(name, "%s.0%2d",  par.SeqName, i_Seq);
	if (i_Seq >= 100)             sprintf(name, "%s.%3d",   par.SeqName, i_Seq);

	stream=fopen(name, "w");
	
	for(idom=0; idom<NDOM; idom++)
		fprintf(stream, "%d \t %d \t ns_and_nt_domain_%1d\n", par.ns[idom], par.nt[idom],idom);
	
	fprintf(stream, "\n");
	fprintf(stream, "%20.15e \t x0\n",    X0FACT*par.xo);
	fprintf(stream, "%20.15e \t x1\n\n",  X1FACT*par.xi);
	print_eos_parameters(   par, stream);
	for(idom=0; idom<NDOM; idom++){
		ns=par.ns[idom];		Ns=ns-1;
		nt=par.nt[idom];		Nt=nt-1;
		for(ipot=0; ipot<NPOT; ipot++){
			for(j=0;j<nt;j++) {
				for(i=0;i<ns;i++){
					n_v = Index(par, idom,ipot,i,j);
					n   = par.n_of_n_v[n_v];
					if(n>=0) fprintf(stream, "%20.15e \t (idom,ipot,i,j)=(%d,%d,%d,%d)\n", 
						X[n], idom, ipot, i, j);
				}
			}
		}
	}
	fprintf(stream, "\n");
	fprintf(stream, "%20.15e \t V0  \n", X[n_2D+0]);
	fprintf(stream, "%20.15e \t Om \n", X[n_2D+1]);
//	fprintf(stream, "%20.15e \t eps \n", X[n_2D+2]);

	fclose(stream);
}
//   -----------------------------------------------
void print_grid(parameters par, derivs_2D v, derivs_1D w, derivs_1D G)
{
	FILE *rho_zeta_values, *output;
	int nx=par.nx, ny=par.ny, i, j, k, l, idom, ns, nt, Ns, Nt, ipot, indx_R;
	double Laplace[NPOT], grad_grad[NPOT*NPOT], u, ux, uy, B, Bx, By, om, omx, omy, mu, mux, muy, vel, vel2,
			e2V, Y, dummy1, dummy2, h, p, eps, epsB,
			xmin=par.xmin, ymin=par.ymin, xmax=par.xmax, ymax=par.ymax,
			***pot[NPOT], ***pots[NPOT], ***pott[NPOT],
			s, t, x, y, a, sa, b, sb, ss, tt;
	coordinate xx, yy;
	derivs_2D V;
	derivs_1D potw[NPOT], Gcheb;
	st st_pair;

	allocate_derivs_2D(&V, NPOT);
	allocate_derivs_1D(&Gcheb, par.ns[NDOM-1]);
	fill0_derivs_1D(Gcheb, par.ns[NDOM-1]);
	for (i=0; i<NPOT; i++){
		allocate_derivs_1D(&potw[i], par.ns[NDOM-1]);
		fill0_derivs_1D(potw[i], par.ns[NDOM-1]);
	}


	ns = par.ns[NDOM-1];
	for (i=0; i<ns; i++) Gcheb.d0[i] = G.d0[i];
	chebft_Extremes(Gcheb.d0, ns, 0); // create Chebyshev coefficients for G
	chder(0., 1., Gcheb.d0, Gcheb.d1,  ns);
	chder(0., 1., Gcheb.d1, Gcheb.d11,  ns);

// two-dimensional fields are defined for the potentials in each domain
	ns=-1; nt=-1; // all ns and nt are equal
	for(idom=0; idom<NDOM; idom++){
		if (par.ns[idom]>ns) ns=par.ns[idom];
		if (par.nt[idom]>nt) nt=par.nt[idom];
	}
	Ns=ns-1; Nt=nt-1;
	for(ipot=0; ipot<NPOT; ipot++){
		pot[ipot]=d3tensor(0,NDOM-1,0,ns,0,nt);
		pots[ipot]=d3tensor(0,NDOM-1,0,ns,0,nt);
		pott[ipot]=d3tensor(0,NDOM-1,0,ns,0,nt);
	}

	for(ipot=0; ipot<NPOT; ipot++) fill0_dtensor(pot[ipot], 0, NDOM-1, 0, ns, 0, nt);


	for(idom=0; idom<NDOM; idom++){
		ns=par.ns[idom]; Ns=ns-1;
		nt=par.nt[idom]; Nt=nt-1;
		for(i=0; i<ns; i++){
			a=Pih*i/Ns, sa=sin(a), s=sa*sa;
			if (idom==NDOM-1){
				for(ipot=0; ipot<NPOT; ipot ++){
					indx_R = Index_Ring(par,ipot,i);
					potw[ipot].d0[i] = w.d0[indx_R];
					potw[ipot].d1[i] = w.d1[indx_R];
					potw[ipot].d11[i] = w.d11[indx_R];
				}
			}
			for(j=0; j<nt; j++){
				b=Pih*j/Nt, sb=sin(b), t=sb*sb;
				coord_st_To_xy(par, idom, i, j, s, t, G, &xx, &yy);
				v_To_V(par, idom, i, j, v, V);
				for(ipot=0; ipot<NPOT; ipot++){
					pot[ipot][idom][i][j]=V.d0[ipot];
					pots[ipot][idom][i][j]=V.d1[ipot];
					pott[ipot][idom][i][j]=V.d2[ipot];
				}
			}
		}
		for(ipot=0; ipot<NPOT; ipot++){
			if (idom==NDOM-1){
				chebft_Extremes(potw[ipot].d0, ns,0);
				chebft_Extremes(potw[ipot].d1, ns,0);
				chebft_Extremes(potw[ipot].d11, ns,0);
			}
			chebft_Extremes_2D(pot[ipot][idom], ns, nt);
			chebft_Extremes_2D(pots[ipot][idom], ns, nt);
			chebft_Extremes_2D(pott[ipot][idom], ns, nt);
		}
	}

	output = fopen("PhysQuant_CylCoord.dat", "w");

	fprintf(output,"#  (1):  x,  (2):    y,  \n");
	fprintf(output,"#  (3):  u,  (4):  u_x,  (5): u_y, \n");
	fprintf(output,"#  (6):  B,  (7):  B_x,  (8): B_y, \n");
	fprintf(output,"#  (9): om, (10): om_x, (11): om_y,\n");
	fprintf(output,"# (12): alpha, (13): alpha_x, (14): alpha_y \n");
	fprintf(output,"# (15):  h, (16):    p\n\n\n");

	printf("Calculation of physical quantities in the cylindrical Box. Please wait ... \n");

	for(k=0; k<=nx; k++){
		x = xmin + (xmax-xmin)*(1.*k/nx)*(1.*k/nx);
		for(l=0; l<=ny; l++){
			y = ymin + (ymax-ymin)*(1.*l/ny)*(1.*l/ny);
//			st_pair = coord_xy_To_st(par, G, x, y);
			st_pair = coord_xy_To_st(par, G, Gcheb, x, y);
			idom = st_pair.idom;
			ss = st_pair.s;
			tt= st_pair.t;
			ns = par.ns[idom]; Ns=ns-1;
			nt = par.nt[idom]; Nt=nt-1;
			coord_st_To_xy_all_st(par, idom, i, j, ss, tt, Gcheb, &xx, &yy);
			for(ipot=0; ipot<NPOT; ipot++){
				if (idom<NDOM-1){
					V.d0[ipot] = chebevxy(0., 1., 0., 1., pot[ipot][idom], ns, nt, ss, tt);
					V.d1[ipot] = chebevxy(0., 1., 0., 1., pots[ipot][idom], ns, nt, ss, tt);
					V.d2[ipot] = chebevxy(0., 1., 0., 1., pott[ipot][idom], ns, nt, ss, tt);
				}
				else{
					V.d0[ipot]  = chebev(0., 1., potw[ipot].d0, ns, ss)
						+ yy.d2*chebevxy(0., 1., 0., 1., pot[ipot][idom], ns, nt, ss, tt);
					V.d1[ipot]  = chebev(0., 1., potw[ipot].d1, ns, ss)
						+ yy.d12*chebevxy(0., 1., 0., 1., pot[ipot][idom], ns, nt, ss, tt)
						+ yy.d2*chebevxy(0., 1., 0., 1., pots[ipot][idom], ns, nt, ss, tt);
					V.d2[ipot]  = yy.d2*chebevxy(0., 1., 0., 1., pott[ipot][idom], ns, nt, ss, tt);
				}
			}

			if (fabs(xx.d0-x)>1.e-6) printf("domain=%d delta_x=%e x_new=%e x=%e y=%e s=%5.5e t=%5.5e\n",
				idom,x-xx.d0,xx.d0,x,y,ss,tt);
			if (fabs(yy.d0-y)>1.e-6) printf("domain=%d delta_y=%e y_new=%e x=%e y=%e s=%5.5e t=%5.5e\n",
				idom,y-yy.d0,yy.d0,x,y,ss,tt);
			pot_st_To_xy(par, xx, yy, V);
			u   = V.d0[0];
			B   = V.d0[1];
			om  = V.d0[2];
			mu  = V.d0[3];
			ux  = V.d1[0];
			uy  = V.d2[0];
			Bx  = V.d1[1];
			By  = V.d2[1];
			omx  = V.d1[2];
			omy  = V.d2[2];
			mux  = V.d1[3];
			muy  = V.d2[3];
			if (idom!=NDOM-1){
				h=p=eps=epsB=0.;
			}
			else{
				vel = sqrt(x)/B*exp(-2.*u)*(par.Om-om);
				vel2= vel*vel;
				e2V = B*B*exp(2.*u)*(1.-vel2);
				h   = par.V0 - 0.5*log(e2V);  // h    = V0 - V - B_Om(Y)
				eos_quantities(par, h, &p, &eps, &epsB);
			}

			fprintf(output,
				"%20.15e \t %20.15e \t %20.15e \t %20.15e \t %20.15e \t %20.15e \t %20.15e \t %20.15e \t %20.15e \t %20.15e \t %20.15e \t %20.15e \t %20.15e \t %20.15e \t %20.15e \t %20.15e\n",
				x, y, u, ux, uy, B, Bx, By, om, omx, omy, mu, mux, muy, exp(h), p);
		}
		printf("%6.1lf%% done.\r", 100.*k/nx); 
		fflush(0);
		fprintf(output,"\n");
	}
	fclose(output);
	printf("\n");
	printf("Please perform `gnuplot PhysQuant_CylCoord.plt' \n\n");
	printf("\n");


	free_derivs_2D(&V, NPOT);
	free_derivs_1D(&Gcheb, par.ns[NDOM-1]);
	for (i=0; i<NPOT; i++) free_derivs_1D(&potw[i], par.ns[NDOM-1]);
	for(ipot=0; ipot<NPOT; ipot++){
		free_d3tensor(pot[ipot],0,NDOM-1,0,ns,0,nt);
		free_d3tensor(pots[ipot],0,NDOM-1,0,ns,0,nt);
		free_d3tensor(pott[ipot],0,NDOM-1,0,ns,0,nt);
	}
}
//   -----------------------------------------------
void print_grid_old(parameters par, derivs_2D v, derivs_1D w, derivs_1D G)
{
	FILE *rho_zeta_values, *output;
	int nx=par.nx, ny=par.ny, i, j, k, l, idom, ns, nt, Ns, Nt, ipot;
	double Laplace[NPOT], grad_grad[NPOT*NPOT], nu, vel, vel2,
			e2V, Y, dummy1, dummy2, h, p, eps, epsB,
			xmin=par.xmin, ymin=par.ymin, xmax=par.xmax, ymax=par.ymax,
			***pot[NPOT], ***potx[NPOT], ***poty[NPOT], potval[NPOT], potvalx[NPOT], potvaly[NPOT],
			s, t, x, y, a, sa, b, sb, ss, tt;
	coordinate xx, yy;
	derivs_2D V;
	st st_pair;

	allocate_derivs_2D(&V, NPOT);

// two-dimensional fields are defined for the potentials in each domain
	ns=par.ns[3]; nt=par.nt[3]; Ns=ns-1; Nt=nt-1; // all ns and nt are equal

	for(ipot=0; ipot<NPOT; ipot++){
		pot[ipot]=d3tensor(0,NDOM-1,0,ns,0,nt);
		potx[ipot]=d3tensor(0,NDOM-1,0,ns,0,nt);
		poty[ipot]=d3tensor(0,NDOM-1,0,ns,0,nt);
	}

	for(idom=0; idom<NDOM; idom++){
		for(i=0; i<ns; i++){
			a=Pih*i/Ns, sa=sin(a), s=sa*sa;
			for(j=0; j<nt; j++){
				b=Pih*j/Nt, sb=sin(b), t=sb*sb;
				if (idom==NDOM-1 && (i==0  || i==ns-1)) {// take values for the inner and outer radii from neighbouring domain
					coord_st_To_xy(par, NDOM-2, i, 0, s, 0., G, &xx, &yy);
					NonLinPotentials(par, NDOM-2, i, 0, s, 0., xx, yy, v, w, G, V, Laplace, grad_grad);
				}
				else{
					coord_st_To_xy(par, idom, i, j, s, t, G, &xx, &yy);
					NonLinPotentials(par, idom, i, j, s, t, xx, yy, v, w, G, V, Laplace, grad_grad);
				}
				for(ipot=0; ipot<NPOT; ipot++) pot[ipot][idom][i][j]=V.d0[ipot];
				if (idom==0 && i==0){
					for(ipot=0; ipot<NPOT; ipot++){
						potx[ipot][idom][i][j]=0.;
						poty[ipot][idom][i][j]=0.;
					}
				}
				else{
					for(ipot=0; ipot<NPOT; ipot++){
						potx[ipot][idom][i][j]=V.d1[ipot];
						poty[ipot][idom][i][j]=V.d2[ipot];
					}
				}
			}
		}
		for(ipot=0; ipot<NPOT; ipot++){
			chebft_Extremes_2D(pot[ipot][idom], ns, nt);
			chebft_Extremes_2D(potx[ipot][idom], ns, nt);
			chebft_Extremes_2D(poty[ipot][idom], ns, nt);
		}
	}

	output = fopen("PhysQuant_CylCoord.dat", "w");

	fprintf(output,"#  (1):  x,  (2):    y,  \n");
	fprintf(output,"#  (3):  u,  (4):  u_x,  (5): u_y, \n");
	fprintf(output,"#  (6):  B,  (7):  B_x,  (8): B_y, \n");
	fprintf(output,"#  (9): om, (10): om_x, (11): om_y,\n");
	fprintf(output,"# (12): alpha, (13): alpha_x, (14): alpha_y \n");
	fprintf(output,"# (15):  h, (16):    p, (17): eps\n\n\n");

	printf("Calculation of physical quantities in the cylindrical Box. Please wait ... \n");

	for(k=0; k<=nx; k++){
		x = xmin + (xmax-xmin)*(1.*k/nx)*(1.*k/nx);
		for(l=0; l<=ny; l++){
			y = ymin + (ymax-ymin)*(1.*l/ny)*(1.*l/ny);
			st_pair = coord_xy_To_st_old(par, G, x, y);
			idom = st_pair.idom;
			ss = st_pair.s;
			tt= st_pair.t;
			ns = par.ns[idom]; Ns=ns-1;
			nt = par.nt[idom]; Nt=nt-1;
			for(ipot=0; ipot<NPOT; ipot++){
				potval[ipot] = chebevxy(0., 1., 0., 1., pot[ipot][idom], ns, nt, ss, tt);
				potvalx[ipot] = chebevxy(0., 1., 0., 1., potx[ipot][idom], ns, nt, ss, tt);
				potvaly[ipot] = chebevxy(0., 1., 0., 1., poty[ipot][idom], ns, nt, ss, tt);
			}
			if (idom!=NDOM-1){
				h=p=eps=epsB=0.;
			}
			else{
				nu  = potval[0] + log(potval[1]);
				vel = sqrt(x)*potval[1]*exp(-2.*nu)*(par.Om-potval[2]);
				vel2= vel*vel;
				e2V = exp(2.*nu)*(1.-vel2);
				h   = par.V0 - 0.5*log(e2V);  // h    = V0 - V - B_Om(Y)
				eos_quantities(par, h, &p, &eps, &epsB);
			}

			fprintf(output,
				"%20.15e \t %20.15e \t %20.15e \t %20.15e \t %20.15e \t %20.15e \t %20.15e \t %20.15e \t %20.15e \t %20.15e \t %20.15e \t %20.15e \t %20.15e \t %20.15e \t %20.15e \t %20.15e\n",
				x, y, potval[0], potvalx[0], potvaly[0], potval[1], potvalx[1], potvaly[1],
				potval[2], potvalx[2], potvaly[2], potval[3], potvalx[3], potvaly[3], exp(h), p);
		}
		printf("%6.1lf%% done.\r", 100.*k/nx); 
		fflush(0);
		fprintf(output,"\n");
	}
	fclose(output);
	printf("\n");
	printf("Please perform `gnuplot PhysQuant_CylCoord.plt' \n\n");
	printf("\n");


	free_derivs_2D(&V, NPOT);
	for(ipot=0; ipot<NPOT; ipot++){
		free_d3tensor(pot[ipot],0,NDOM-1,0,ns,0,nt);
		free_d3tensor(potx[ipot],0,NDOM-1,0,ns,0,nt);
		free_d3tensor(poty[ipot],0,NDOM-1,0,ns,0,nt);
	}

}
//   -----------------------------------------------
void print_grid_older(parameters par, derivs_2D v, derivs_1D w, derivs_1D G)
{
	FILE *rho_zeta_values, *output;
	int nx=par.nx, ny=par.ny, i, j, k, l, idom, ns, nt, Ns, Nt, ipot;
	double Laplace[NPOT], grad_grad[NPOT*NPOT], nu, vel, vel2,
			e2V, Y, dummy1, dummy2, h, p, eps, epsB,
			xmin=par.xmin, ymin=par.ymin, xmax=par.xmax, ymax=par.ymax,
			u, ux, uy, B, Bx, By, om, omx, omy, mu, mux, muy,
			cheb_ut[NMAX], cheb_u[NMAX], cheb_uxt[NMAX], cheb_ux[NMAX], cheb_uyt[NMAX], cheb_uy[NMAX],
			cheb_Bt[NMAX], cheb_B[NMAX], cheb_Bxt[NMAX], cheb_Bx[NMAX], cheb_Byt[NMAX], cheb_By[NMAX],
			cheb_omt[NMAX], cheb_om[NMAX], cheb_omxt[NMAX], cheb_omx[NMAX], cheb_omyt[NMAX], cheb_omy[NMAX],
			cheb_mut[NMAX], cheb_mu[NMAX], cheb_muxt[NMAX], cheb_mux[NMAX], cheb_muyt[NMAX], cheb_muy[NMAX],
			s, t, x, y, a, sa, b, sb, ss, tt;
	coordinate xx, yy;
	derivs_2D V;
	st st_pair;

	allocate_derivs_2D(&V, NPOT);

// two-dimensional fields are defined for the potentials in each domain
	output = fopen("PhysQuant_CylCoord.dat", "w");

	fprintf(output,"#  (1):  x,  (2):    y,  \n");
	fprintf(output,"#  (3):  u,  (4):  u_x,  (5): u_y, \n");
	fprintf(output,"#  (6):  B,  (7):  B_x,  (8): B_y, \n");
	fprintf(output,"#  (9): om, (10): om_x, (11): om_y,\n");
	fprintf(output,"# (12): alpha, (13): alpha_x, (14): alpha_y \n");
	fprintf(output,"# (15):  h, (16):    p\n\n\n");

	printf("Calculation of physical quantities in the cylindrical Box. Please wait ... \n");

	for(k=0; k<=nx; k++){
		x = xmin + (xmax-xmin)*(1.*k/nx)*(1.*k/nx);
		for(l=0; l<=ny; l++){
			y = ymin + (ymax-ymin)*(1.*l/ny)*(1.*l/ny);
			st_pair = coord_xy_To_st_old(par, G, x, y);
			idom = st_pair.idom;
			ss = st_pair.s;
			tt= st_pair.t;
			ns = par.ns[idom]; Ns=ns-1;
			nt = par.nt[idom]; Nt=nt-1;
			for(i=0; i<ns; i++){
				a=Pih*i/Ns, sa=sin(a), s=sa*sa;
				for(j=0; j<nt; j++){
					b=Pih*j/Nt, sb=sin(b), t=sb*sb;
					if (idom==NDOM-1 && (i==0  || i==ns-1)) {// take values for the inner and outer radii from neighbouring domain
						coord_st_To_xy(par, NDOM-2, i, 0, s, 0., G, &xx, &yy);
						NonLinPotentials(par, NDOM-2, i, 0, s, 0., xx, yy, v, w, G, V, Laplace, grad_grad);
					}
					else{
						coord_st_To_xy(par, idom, i, j, s, t, G, &xx, &yy);
						NonLinPotentials(par, idom, i, j, s, t, xx, yy, v, w, G, V, Laplace, grad_grad);
					}
					cheb_ut[j]  = V.d0[0]; cheb_Bt[j]  = V.d0[1]; cheb_omt[j]  = V.d0[2]; cheb_mut[j]  = V.d0[3];
					if (idom==0 && i==0){
						cheb_uxt[j] = cheb_Bxt[j] = cheb_omxt[j] = cheb_muxt[j] = 0.;
						cheb_uyt[j] = cheb_Byt[j] = cheb_omyt[j] = cheb_muyt[j] = 0.;
					}
					else{
						cheb_uxt[j] = V.d1[0]; cheb_Bxt[j] = V.d1[1]; cheb_omxt[j] = V.d1[2]; cheb_muxt[j] = V.d1[3];
						cheb_uyt[j] = V.d2[0]; cheb_Byt[j] = V.d2[1]; cheb_omyt[j] = V.d2[2]; cheb_muyt[j] = V.d2[3];
					}
				}
				chebft_Extremes(cheb_ut, nt, 0); chebft_Extremes(cheb_uxt, nt, 0); chebft_Extremes(cheb_uyt, nt, 0);
				chebft_Extremes(cheb_Bt, nt, 0); chebft_Extremes(cheb_Bxt, nt, 0); chebft_Extremes(cheb_Byt, nt, 0);
				chebft_Extremes(cheb_omt, nt, 0); chebft_Extremes(cheb_omxt, nt, 0); chebft_Extremes(cheb_omyt, nt, 0);
				chebft_Extremes(cheb_mut, nt, 0); chebft_Extremes(cheb_muxt, nt, 0); chebft_Extremes(cheb_muyt, nt, 0);
				cheb_u[i]  = chebev(0., 1., cheb_ut, nt, tt);
					cheb_ux[i] = chebev(0., 1., cheb_uxt, nt, tt);
						cheb_uy[i] = chebev(0., 1., cheb_uyt, nt, tt);
				cheb_B[i]  = chebev(0., 1., cheb_Bt, nt, tt);
					cheb_Bx[i] = chebev(0., 1., cheb_Bxt, nt, tt);
						cheb_By[i] = chebev(0., 1., cheb_Byt, nt, tt);
				cheb_om[i]  = chebev(0., 1., cheb_omt, nt, tt);
					cheb_omx[i] = chebev(0., 1., cheb_omxt, nt, tt);
						cheb_omy[i] = chebev(0., 1., cheb_omyt, nt, tt);
				cheb_mu[i]  = chebev(0., 1., cheb_mut, nt, tt);
					cheb_mux[i] = chebev(0., 1., cheb_muxt, nt, tt);
						cheb_muy[i] = chebev(0., 1., cheb_muyt, nt, tt);
			}
			chebft_Extremes(cheb_u, ns, 0); chebft_Extremes(cheb_ux, ns, 0); chebft_Extremes(cheb_uy, ns, 0);
			chebft_Extremes(cheb_B, ns, 0); chebft_Extremes(cheb_Bx, ns, 0); chebft_Extremes(cheb_By, ns, 0);
			chebft_Extremes(cheb_om, ns, 0); chebft_Extremes(cheb_omx, ns, 0); chebft_Extremes(cheb_omy, ns, 0);
			chebft_Extremes(cheb_mu, ns, 0); chebft_Extremes(cheb_mux, ns, 0); chebft_Extremes(cheb_muy, ns, 0);
			u  = chebev(0., 1., cheb_u, ns, ss);
			ux = chebev(0., 1., cheb_ux, ns, ss);
			uy = chebev(0., 1., cheb_uy, ns, ss);
			B  = chebev(0., 1., cheb_B, ns, ss);
			Bx = chebev(0., 1., cheb_Bx, ns, ss);
			By = chebev(0., 1., cheb_By, ns, ss);
			om  = chebev(0., 1., cheb_om, ns, ss);
			omx = chebev(0., 1., cheb_omx, ns, ss);
			omy = chebev(0., 1., cheb_omy, ns, ss);
			mu  = chebev(0., 1., cheb_mu, ns, ss);
			mux = chebev(0., 1., cheb_mux, ns, ss);
			muy = chebev(0., 1., cheb_muy, ns, ss);
			if (idom!=NDOM-1){
				h=p=eps=epsB=0.;
			}
			else{
				nu  = u + log(B);
				vel = sqrt(x)*B*exp(-2.*nu)*(par.Om-om);
				vel2= vel*vel;
				e2V = exp(2.*nu)*(1.-vel2);
				h   = par.V0 - 0.5*log(e2V);  // h    = V0 - V - B_Om(Y)
				eos_quantities(par, h, &p, &eps, &epsB);
			}
			fprintf(output,
				"%20.15e \t %20.15e \t %20.15e \t %20.15e \t %20.15e \t %20.15e \t %20.15e \t %20.15e \t %20.15e \t %20.15e \t %20.15e \t %20.15e \t %20.15e \t %20.15e \t %20.15e \t %20.15e \t %20.15e\n",
				x, y, u, ux, uy, B, Bx, By, om, omx, omy, mu, mux, muy, exp(h), p, eps);
		}
	printf("%6.1lf%% done.\r", 100.*k/nx); 
	fflush(0);
	fprintf(output,"\n");
	}
	fclose(output);
	printf("\n");
	printf("Please perform `gnuplot PhysQuant_CylCoord.plt' \n\n");
	printf("\n");

	free_derivs_2D(&V, NPOT);
}
